from p2pd import *

class NetDebugServer(Daemon):
    async def msg_cb(self, msg, client_tup, pipe):
        # Parse HTTP message and handle CORS.
        req = await rest_service(msg, client_tup, pipe)

        # Output JSON response for the API.
        async def get_response():
            # Send 'hello' to a remote client to test it.
            named = ["proto", "dest_addr", "dest_port", "route_ip", "src_port"]
            ip_p = "[0-9a-fA-F:.]"
            p = req.api(f"/hello/((?:udp)|(?:tcp))/([^/]+)/([0-9]+)/({ip_p}+)/([0-9]+)", named)
            if p:
                # Get port and check it's valid.
                dest_port = to_n(p["dest_port"])
                if not valid_port(dest_port):
                    return {
                        "error": 1,
                        "msg": "invalid port for /hello"
                    }

                # Check src port.
                src_port = to_n(p["src_port"])
                if src_port != 0:
                    if not valid_port(src_port):
                        return {
                            "error": 6,
                            "msg": "invalid src port for /hello"
                        }

                # Error thrown and logged in handler if not valid.
                route_ipr = IPRange(p["route_ip"])
                af = route_ipr.af
                route = await self.rp[af].locate(route_ipr).bind(port=src_port)
                if route is None:
                    return {
                        "error": 3,
                        "msg": "this server doesn't control that IP."
                    }

                # Make pipe used to send hello message.
                proto = PROTO_LOOKUP[p["proto"].upper()]
                dest_addr = await Address(p["dest_addr"], dest_port, route).res()
                hello_pipe = await pipe_open(proto, route, dest_addr)
                hello_from = hello_pipe.sock.getsockname()

                # Send hello and close the pipe.
                await hello_pipe.send(b"hello")
                await hello_pipe.close()

                return {
                    "error": 0,
                    "msg": f"sent hello to {p['proto']}:{dest_addr.tup} from {hello_from} route = {route}"
                }

            # Make a TCP connection back to a service to test port forwarding.
            named = ["port"]
            p = req.api(f"/reverse/([0-9]+)", named)
            if p:
                # Get port and check it's valid.
                dest_port = to_n(p["port"])
                if not valid_port(dest_port):
                    return {
                        "error": 4,
                        "msg": "invalid port for /reverse"
                    }

                # Peers address of this connection.
                dest_tup = (pipe.sock.getpeername()[0], dest_port)                
                try:
                    # Resolve dest tuple to address.
                    dest_addr = await Address(*dest_tup, pipe.route, timeout=1).res()

                    # Attempt to make TCP con to dest_port.
                    route = copy.deepcopy(pipe.route)
                    tcp_pipe = await pipe_open(
                        route=await route.bind(),
                        proto=TCP,
                        dest=dest_addr,
                        conf=dict_child({
                            "con_timeout": 1
                        }, NET_CONF) 
                    )

                    # Failure.
                    if tcp_pipe is None:
                        raise Exception("Failed to connect.")

                    # Success if no exception fired.
                    await tcp_pipe.close()
                    return {
                        "error": 0,
                        "msg": f"tcp service {dest_tup} is reachable."
                    }
                except Exception:
                    return {
                        "error": 5,
                        "msg": f"tcp service {dest_tup} is not reachable."
                    }

            # Show information of the connection to the peer.
            p = req.api("/mapping")
            if p:
                sock = pipe.sock
                return {
                    "error": 0,
                    "fd": sock.fileno(),
                    "laddr": sock.getsockname(),
                    "raddr": sock.getpeername(),
                    "route": pipe.route.to_dict(),
                    "if": {
                        "name": pipe.route.interface.name
                    }
                }

            # Fallback and serve all other purposes.
            return {
                "error": 2,
                "msg": "method not implemented"
            }

        # Return JSON response to browser.
        resp = await get_response()
        if resp is not None:
            await send_json(
                resp,
                req,
                client_tup,
                pipe
            )

async def start_net_debug():
    i = await Interface().start()
    net_debug_server = NetDebugServer()
    net_debug_server.set_rp(i.rp)
    await net_debug_server.listen_all(
        [i.rp[IP4], i.rp[IP6]],
        [20000],
        [TCP, UDP]
    )

    while 1:
        await asyncio.sleep(10)

async_test(start_net_debug)


import asyncio
from p2pd.test_init import *
from p2pd.p2p_node import *
from p2pd.p2p_utils import *
import multiprocessing
from decimal import Decimal as Dec

asyncio.set_event_loop_policy(SelectorEventPolicy())
class TestTCPPunch(unittest.IsolatedAsyncioTestCase):
    async def test_map_info_to_their_maps(self):
        map_info = {
            "remote": [5000],
            "reply": [0],
            "local": [10000]
        }

        expected = [
            [ 5000, 0, 10000 ]
        ]

        self.assertEqual(
            map_info_to_their_maps(map_info),
            expected
        )

    async def test_self_punch_no_networking(self):
        # Used to schedule and synchronize the punching.
        # Manager is used to push sockets between processes.
        log(">>> test_self_punch_no_networking")
        sys_clock = SysClock(Dec("0.01"))
        pe = await get_pp_executors()
        #pe2 = await get_pp_executors(workers=2)
        
        if pe is not None:
            qm = multiprocessing.Manager()
        else:
            qm = None
            
        pipe_id = b"pipe_id"

        # Load interfaces is slow AF on Windows.
        # Due to it using powershell + regex, lolz.
        ifs = await load_interfaces()
        interface = ifs[0]
        if IP4 in interface.supported():
            af = IP4
        else:
            af = IP6

        # Pretend we're using an open internet.
        # This is because we're doing a self-test.
        dest = interface.rp[af].routes[0].nic()
        #nat = nat_info(OPEN_INTERNET, delta_info(NA_DELTA, 0))
        nat = nat_info(FULL_CONE, delta_info(PRESERV_DELTA, 0))
        nat["is_concurrent"] = random.choice([True, False])
        for i in ifs:
            i.set_nat(nat)

            # Not needed for a self / LAN test:
            # await interface.load_nat()

        # Used to get external mappings (not needed for self-test.)
        stun_client = STUNClient(interface, af)
        
        # Initiator client -- starts the protocol.
        # Sends the first mapping details.
        i_node_id = b"i_node_id"
        i_client = TCPPunch(interface, ifs, sys_clock, pe, qm)
        i_client.do_state_cleanup = False

        # Recipient client -- receives the first mappings.
        # Provides additional mapping details if possible.
        r_node_id = b"r_node_id"
        r_client = TCPPunch(interface, ifs, sys_clock, pe, qm)
        r_client.do_state_cleanup = False

        # Get punch mode.
        route = interface.route(af)
        dest_addr = await Address(dest, 80, route).res()
        mode = i_client.get_punch_mode(dest_addr)
        self.assertEqual(mode, TCP_PUNCH_SELF)

        # Get initial mappings.
        i_maps, ntp_meet, got_update = await i_client.proto_send_initial_mappings(
            dest_addr=dest,
            dest_nat=nat,
            dest_node_id=r_node_id,
            pipe_id=pipe_id,
            stun_client=stun_client,
            mode=mode
        )

        #print(i_maps, ntp_meet)

        # Receive their mappings.
        r_maps, _, sent_update = await r_client.proto_recv_initial_mappings(
            recv_addr=dest,
            recv_nat=nat,
            recv_node_id=i_node_id,
            pipe_id=pipe_id,
            their_maps=i_maps,
            stun_client=stun_client,
            ntp_meet=ntp_meet,
            mode=mode
        )

        #print(r_maps)

        await i_client.proto_update_recipient_mappings(
            dest_node_id=r_node_id,
            pipe_id=pipe_id,
            their_maps=r_maps,
            stun_client=stun_client
        )

        #print(i_client.state)

        try:
            results = await asyncio.wait_for(
                asyncio.gather(
                    async_wrap_errors(
                        i_client.proto_do_punching(PUNCH_INITIATOR, r_node_id, pipe_id)
                    ),
                    async_wrap_errors(
                        r_client.proto_do_punching(PUNCH_RECIPIENT, i_node_id, pipe_id)
                    )
                ),
                10
            )
        except Exception:
            results = []

        print("Got results = ")
        print(results)

        async def process_results(results):
            if len(results) != 2:
                return False

            msg = b"a test msg"
            for pipe in results:
                if pipe is None:
                    return False

                pipe.subscribe(sub=SUB_ALL)

            for pipe in results:
                await pipe.send(msg, pipe.stream.dest_tup)

            for pipe in results:
                out = await pipe.recv(timeout=3)
                if out != msg:
                    return False

            return True

        status = await process_results(results)
        if status == False:
            print("Self punch test no networking failed.")
            print("Test is optional but noting it here.")


        for pipe in results:
            if pipe is not None:
                await pipe.close()

        for punch_client in [i_client, r_client]:
            await punch_client.close()

        # Try close the multiprocess manager.
        try:
            if qm is not None:
                qm.shutdown()
        except Exception:
            pass

        # Try close the process pool executor.
        try:
            if pe is not None:
                pe.shutdown()
        except Exception:
            pass

if __name__ == '__main__':
    main()


"""
Given an NTP accurate time, this module computes an
approximation of how far off the system clock is
from the NTP time (clock skew.) The algorithm was
taken from gtk-gnutella.

https://github.com/gtk-gnutella/gtk-gnutella/
blob/devel/src/core/clock.c 

Original Python version by... myself! Now with async.

This code asks for many NTP readings which are slow.
It will probably be necessary to initalize this in
an ensure_future that sets related objects when done.
It also doesn't support the Interface object or
Address hence it defaults to the default route.

https://datatracker.ietf.org/doc/html/rfc5905#section-6
"""

import random
from decimal import Decimal as Dec
from .address import *
from .ntp_client import NTPClient
from .settings import *

NTP_RETRY = 2
NTP_TIMEOUT = 2

async def get_ntp(interface, server=None, retry=NTP_RETRY):
    server = server or random.choice(NTP_SERVERS)[0]
    try:
        for _ in range(retry):
            client = NTPClient(interface)
            response = await client.request(server, version=3)
            if response is None:
                continue

            ntp = response.tx_time
            return ntp
    except Exception as e:
        log_exception()
        return None

class SysClock:
    def __init__(self, interface, clock_skew=Dec(0)):
        self.interface = interface
        self.enough_data = 40
        self.min_data = 10
        self.max_sdev = 60
        self.clean_steps = 3
        self.data_points = []
        self.clock_skew = clock_skew
        if self.clock_skew != Dec(0):
            self.clock_skew = Dec(clock_skew)

    async def start(self):
        if not self.clock_skew:
            # Test whether this host has an NTPD.
            """
            'NTP can usually maintain time to within tens of milliseconds over the public Internet, and can achieve better than one millisecond accuracy in local area networks under ideal conditions.'
            Plenty accurate for hole punching.
            """
            # NTPD listens on all interfaces so
            # the LAN IP doesn't matter.
            server = None
            local_ip = "localhost"
            ntp_ret = await async_wrap_errors(
                get_ntp(
                    self.interface,
                    local_ip
                ),
                timeout=NTP_RETRY * NTP_TIMEOUT
            )

            if ntp_ret is not None:
                log("> clockskew using local ntp daemon")
                server = local_ip

            # Calculate clock skew.
            for i in range(0, 3):
                if self.clock_skew == Dec(0):
                    await self.collect_data_points(server=server)
                    if not len(self.data_points):
                        continue

                    self.clock_skew = self.calculate_clock_skew()
                else:
                    break

            # Raise exception.
            if self.clock_skew == Dec(0):
                raise Exception("Could not get clock skew.")
    
        return self

    def __await__(self):
        return self.start().__await__()

    def __str__(self):
        return '{0:f}'.format(self.clock_skew)

    def __repr__(self):
        return f'Dec("{str(self)}")'

    def time(self):
        return Dec(timestamp(1)) - self.clock_skew

    async def collect_data_points(self, server=None):
        async def get_clock_skew():
            ntp_ret = await async_wrap_errors(
                get_ntp(self.interface, server=server),
                timeout=NTP_RETRY * NTP_TIMEOUT
            )

            if ntp_ret is None:
                return None

            return Dec(timestamp(1)) - Dec(ntp_ret)

        tasks = []
        for _ in range(0, self.enough_data + 10):
            tasks.append(
                get_clock_skew()
            )

        results = await asyncio.gather(*tasks)
        results = strip_none(results)
        self.data_points += results

    def statx_n(self, data_points):
        return len(data_points)

    def statx_avg(self, data_points):
        total = Dec("0")
        n = self.statx_n(data_points)
        for i in range(0, n):
            total += data_points[i]

        return total / Dec(n)

    def statx_sdev(self, data_points):
        def _ss(data):
            # Return sum of square deviations
            # of sequence data.
            c = self.statx_avg(data)
            return sum( (x - c ) ** 2 for x in data )

        def pstdev(data):
            # Calculates the population standard deviation.
            n = len(data)
            if n < 2:
                raise ValueError('variance requires at least two data points')

            ss = _ss(data)
            pvar = ss / n  # the population variance

            return pvar ** Dec("0.5")

        return pstdev(data_points)

    def calculate_clock_skew(self):
        """
        Computer average and standard deviation
        using all the data points.
        """
        n = self.statx_n(self.data_points)

        """
        Required to be able to compute the standard
        deviation.
        """
        if n < 1:
            log("n < 1 in cal clock skew")
            return Dec("0")

        avg = self.statx_avg(self.data_points)
        sdev = self.statx_sdev(self.data_points)

        """
        Incrementally remove aberration points.
        """
        for k in range(0, self.clean_steps):
            """
            Remove aberration points: keep only
            the sigma range around the average.
            """
            min_val = avg - sdev
            max_val = avg + sdev

            cleaned_data_points = []
            for i in range(0, n):
                v = self.data_points[i]
                if v < min_val or v > max_val:
                    continue

                cleaned_data_points.append(v)

            self.data_points = cleaned_data_points[:]

            """
            Recompute the new average using the
            "sound" points we kept.
            """
            n = self.statx_n(self.data_points)

            """
            Not enough data to compute standard
            deviation.
            """
            if n < 2:
                break

            avg = self.statx_avg(self.data_points)
            sdev = self.statx_sdev(self.data_points)
            if sdev <= self.max_sdev or n < self.min_data:
                break

        """
        If standard deviation is too large still, we
        cannot update our clock. Collect more points.

        If we don't have a minimum amount of data,
        don't attempt the update yet, continue collecting.
        """
        if sdev > self.max_sdev or n < self.min_data:
            return Dec("0")

        return avg

    def to_dict(self):
        return {
            "if": self.interface.to_dict(),
            "clock_skew": self.clock_skew
        }

    @staticmethod
    def from_dict(d):
        from p2pd.interface import Interface
        i = Interface.from_dict(d["if"])
        x = SysClock(interface=i, clock_skew=d["clock_skew"])
        return x

    # Pickle.
    def __getstate__(self):
        return self.to_dict()

    # Unpickle.
    def __setstate__(self, state):
        o = self.from_dict(state)
        self.__dict__ = o.__dict__

async def test_clock_skew(): # pragma: no cover
    from p2pd.interface import init_p2pd, Interface
    interface = await Interface().start_local()
    s = await SysClock(interface=interface)
    print(repr(s))


    return
    f = []
    for i in range(len(NTP_SERVERS)):
        out = await get_ntp(interface, server=NTP_SERVERS[i][0])
        if out is None:
            f.append(i)

    print(f)
        
if __name__ == "__main__":
    #sys_clock = SysClock()
    #print(sys_clock.clock_skew)
    async_test(test_clock_skew)


    # print(get_ntp())
    # print(get_ntp())

    """
    print(sys_clock.time())
    print()
    print(get_ntp())
    print(sys_clock.time())
    print()
    print(get_ntp())
    print(sys_clock.time())
    """



from .net import *
from .ip_range import *

# Netifaces apparently doesn't use their own values...
def af_to_netiface(af):
    if af == IP4:
        return int(IP4)
        return netifaces.AF_INET
        
    if af == IP6:
        return int(IP6)
        return netifaces.AF_INET6

    return af

def netiface_to_af(af, netifaces):
    if af == netifaces.AF_INET:
        return IP4

    if af == netifaces.AF_INET6:
        return IP6

    return af

def is_af_routable(af, netifaces):
    af = af_to_netiface(af)
    return af in netifaces.gateways()

def get_mac_address(name, netifaces):
    return netifaces.ifaddresses(name)[netifaces.AF_LINK][0]["addr"]
    
# Note: Discards subnet for single addresses.
async def netiface_addr_to_ipr(af, info, interface, loop, skip_bind_test):
    # Some interfaces might not have valid information set.
    if "addr" not in info:
        return None
    if "netmask" not in info:
        return None
    
    nic_ipr = IPRange(info["addr"], info["netmask"])

    """
    Some operating systems incorrectly list the netmask for
    single IPs. This is a patch that discards the netmask if
    a host portion is detected. Ranges should have all zeros
    for the host portion.
    """
    if nic_ipr.i_host:
        nic_ipr = IPRange(info["addr"])

    """
    If a range is detected test that the range is valid by
    trying to bind on the first and last address.
    """
    if not skip_bind_test and not nic_ipr.i_host:
        invalid_subnet = False
        for host_index in [0, -1]:
            ip_obj = nic_ipr[host_index]
            bind_ip = str(ip_obj)
            if af == IP6:
                bind_ip = ip6_patch_bind_ip(ip_obj, bind_ip, interface)

            s = socket.socket(af, TCP)
            try:
                addr_infos = await loop.getaddrinfo(
                    bind_ip,
                    0
                )
                s.bind(addr_infos[0][4])
            except Exception:
                log_exception()
                log(f"af = {af}, bind_ip = {bind_ip}")
                log(f"{addr_infos[0][4]}")
                log(f"{s}")
                log(">get routes invalid subnet for {}".format(str(nic_ipr)))
                invalid_subnet = True
            finally:
                s.close()
                break

        # Don't add this address to any route.
        if invalid_subnet:
            return None

    return nic_ipr

async def get_nic_private_ips(interface, af, netifaces, loop=None):
    loop = loop or asyncio.get_event_loop()
    nic_iprs = []
    if_name = interface.name
    if_addresses = netifaces.ifaddresses(if_name)
    if af not in if_addresses:
        return nic_iprs

    for info in if_addresses[af]:
        nic_ipr = await netiface_addr_to_ipr(af, info, interface, loop, skip_bind_test=0)
        if nic_ipr is None:
            continue

        if nic_ipr.is_public and af == IP6:
            continue

        nic_iprs.append(nic_ipr)

    return nic_iprs

def netiface_gateways(netifaces, get_interface_type, preference=AF_ANY):
    gws = netifaces.gateways()
    gateway = None
    iface = None
    
    # Netifaces may not always find the default gateway.
    # Use first interface that get_interface_type finds.
    if gws["default"] == {}:
        # Create a list of address families to check.
        if preference == AF_ANY:
            afs = VALID_AFS
        else:
            afs = [preference]
            
        # Check the address families of related GWs.
        for af in afs:
            # No entry for AF found in GW info.
            if af not in gws:
                continue
                
            # Check that there is info to check.
            if not len(gws[af]):
                continue
                
            # Check the interface name.
            for net_info in gws[af]:
                # Af already set.
                if af in gws["default"]:
                    continue
                
                # Try to determine interface type from name.
                if_name = net_info[1]
                if_type = get_interface_type(if_name)
                
                # Unknown / bad interface type.
                if if_type == INTERFACE_UNKNOWN:
                    continue
                    
                # Use this interface GW info as the default.
                gws["default"][af] = net_info
                break
                
    return gws




from decimal import Decimal as Dec
from p2pd.test_init import *
from p2pd.utils import async_test
from p2pd.nat import *
from p2pd.p2p_node import *
from p2pd.p2p_utils import *
from p2pd.stun_client import STUNClient

if __name__ == "__main__":
    async def test_p2p_node():
        # Output the nodes p2p address.
        signal_offsets = [0, 3]
        node = await start_p2p_node(
            # Used for certain tests.
            node_id=b"p2pd_test_node",

            # Fixed interface dict for faster startup.
            ifs=P2PD_IFS,

            # Fixed list of MQTT server offsets to use.
            signal_offsets=signal_offsets,
        )
        print(node.addr_bytes)

        # Prevent process from exiting.
        while 1:
            await asyncio.sleep(1)

    async_test(test_p2p_node)

"""
rm -rf program.log; while true; do python3 main.py; sleep 1; done
"""


"""
    async def test_irc_dns():
        # The 'scanner' was basically this initially.
        # might as well get it working again so it can be reused.
        tasks = []
        for server in IRC_SERVERS1:
            task = async_wrap_errors(IRCDNS(server).start(i))
            tasks.append(task)

        out = await asyncio.gather(*tasks)
        print(out)
        out = strip_none(out)
        print(out)
        return
"""

import pprint
from p2pd.test_init import *
from p2pd import *

IRC_SEED = b"123" * 30
IRC_SERV_INFO = {
    'domain': 'example.com',
    'afs': [IP4, IP6],

    # 6 nov 2000
    'creation': 975848400,

    'nick_serv': ["password", "email"],

    "ip": {
        IP4: "93.184.216.34",
        IP6: "2606:2800:220:1:248:1893:25c8:1946]"
    },

    'chan_len': 50,
    'chan_no': 60,
    'topic_len': 300,
    'chan_topics': 'a-zA-Z0-9all specials unicode (smilies tested)'
}

IRC_S = IRCSession(IRC_SERV_INFO, IRC_SEED)

IRC_TEST_SERVERS_SEVEN = [
    {"domain": "a", "chan_expiry": 14, "nick_expiry": 14},
    {"domain": "b", "chan_expiry": 14, "nick_expiry": 14},
    {"domain": "c", "chan_expiry": 14, "nick_expiry": 14},
    {"domain": "d", "chan_expiry": 14, "nick_expiry": 14},
    {"domain": "e", "chan_expiry": 14, "nick_expiry": 14},
    {"domain": "f", "chan_expiry": 14, "nick_expiry": 14},
    {"domain": "g", "chan_expiry": 14, "nick_expiry": 14},
]

class MockIRCChan(IRCChan):
    async def set_topic(self, topic):
        self.pending_topic = topic

class MockIRCSession(IRCSession):
    async def start(self, i):
        if self.db is not None:
            last_started_key = self.db_key("last_started")
            await self.db.put(last_started_key, time.time())

            # Save user details if needed.
            nick_key = self.db_key("nick")
            await self.db.put(nick_key, {
                "domain": self.irc_server,
                "nick": self.nick,
                "username": self.username,
                "user_pass": self.user_pass,
                "email": self.email,
                "last_refresh": time.time()
            })

        self.started.set_result(True)

    async def is_chan_registered(self, chan_name):
        if chan_name not in self.chan_registered:
            return False
        else:
            return self.nick
    
    async def register_chan(self, chan_name):
        self.chan_registered[chan_name] = True

    async def get_chan_topic(self, chan_name):
        return self.chans[chan_name].pending_topic


# python -m unittest test_irc_dns.TestIRCDNS.
class TestIRCDNS(unittest.IsolatedAsyncioTestCase):
    async def test_proto_ping(self):
        msg = IRCMsg(cmd="PING", param="31337")
        resp = irc_proto(None, msg)
        assert(resp.pack() == b"PONG 31337\r\n")

        msg = IRCMsg(cmd="PING", suffix="31337")
        resp = irc_proto(None, msg)
        assert(resp.pack() == b"PONG :31337\r\n")

    async def test_proto_ctcp_version(self):
        msg = IRCMsg(
            cmd="PRIVMSG",
            prefix="user!ident@host",
            suffix="\x01VERSION\x01"
        )

        resp = irc_proto(None, msg)
        expected = to_b(f"PRIVMSG user :\x01VERSION {IRC_VERSION}\x01\r\n")
        assert(resp.pack() == expected)

    async def test_proto_is_chan_reg(self):
        chan_founder = "chan_founder"
        chan_name = "#test-wrwerEWER342"
        vectors = [
            [f"channel {chan_name} isn't", False],
            [f"channel {chan_name} is not", False],
            [f"information for {chan_name}", chan_founder],
            [f"information on {chan_name}", chan_founder],
            [f"channel {chan_name} is registered", chan_founder]
        ]

        for vector in vectors:
            IRC_S.chan_infos[chan_name] = asyncio.Future()

            status, expected = vector
            msg = IRCMsg(
                cmd="NOTICE",
                param="your_nick",
                suffix=status
            )

            irc_proto(IRC_S, msg)

            if expected:
                msg = IRCMsg(
                    cmd="NOTICE",
                    param="your_nick",
                    suffix=f"Founder : {chan_founder}"
                )

                irc_proto(IRC_S, msg)

            assert(IRC_S.chan_infos[chan_name].result() == expected)

    async def test_proto_get_topic(self):
        chan_name = "#test-CAJANS324"
        chan_topic = B92_CHARSET + " " + B92_CHARSET
        msg = IRCMsg(
            cmd="332",
            param=f"nick {chan_name}",
            suffix=chan_topic
        )

        IRC_S.chan_topics[chan_name] = asyncio.Future()
        irc_proto(IRC_S, msg)
        assert(IRC_S.chan_topics[chan_name].result() == chan_topic)

    async def test_irc_extract_msg(self):
        vectors = [
            [
                "CMD value\r\n",
                IRCMsg(cmd="CMD", param="value")
            ],
            [
                ":test CMD v\r\n",
                IRCMsg(
                    prefix="test",
                    cmd="CMD",
                    param="v"
                )
            ],
            [
                "CMD v :suffix\r\n",
                IRCMsg(
                    suffix="suffix",
                    cmd="CMD",
                    param="v"
                )
            ],
            [
                ":prefix-part CMD v long param :suffix part\r\n",
                IRCMsg(
                    prefix="prefix-part",
                    cmd="CMD",
                    param="v long param",
                    suffix="suffix part"
                )
            ],
            [
                ":prefix-part   CMD  v long param :suffix part\r\n",
                IRCMsg(
                    prefix="prefix-part",
                    cmd="CMD",
                    param="v long param",
                    suffix="suffix part"
                )
            ],
        ]

        for vector in vectors:
            buf, expected = vector
            got, _ = irc_extract_msgs(buf)
            got = got[0]
            assert(got == expected)

    async def test_irc_extract_sender(self):
        vectors = [
            [
                "nickaAWE234 asd",
                {
                    "nick": "nickaAWE234 asd",
                    "user": "",
                    "host": ""
                }
            ],
            [
                "nickaAWE234 asd@hostwe12 S",
                {
                    "nick": "nickaAWE234 asd",
                    "user": "",
                    "host": "hostwe12 S"
                }
            ],
            [
                "nickaAWE234 asd!user sF S",
                {
                    "nick": "nickaAWE234 asd",
                    "user": "user sF S",
                    "host": ""
                }
            ],
            [
                "nickaAWE234 asd!user sF S@HoST N",
                {
                    "nick": "nickaAWE234 asd",
                    "user": "user sF S",
                    "host": "HoST N"
                }
            ]
        ]

        for vector in vectors:
            buf, expected = vector
            got = irc_extract_sender(buf)
            assert(got == expected)

    async def test_start_n(self):
        servers = [
            {"domain": "a"},
            {"domain": "b"},
            {"domain": "c"},
            {"domain": "d"},
            {"domain": "e"},
        ]

        interface = None
        ircdns = await IRCDNS(
            i=interface,
            seed=b"1" + IRC_SEED,
            clsSess=MockIRCSession,
            clsChan=MockIRCChan,
            servers=servers,
            executor=executor,
            do_shuffle=False
        ).start()

        # Sanity checks on length.

        # Should continue.
        await ircdns.start_n(len(servers) - 1)
        assert(ircdns.p_sessions_next == len(servers) - 1)

        await ircdns.start_n(1)
        assert(ircdns.p_sessions_next == len(servers))

        #cd projects/p2pd/tests
        #  python -m unittest test_irc_dns.TestIRCDNS.test_start_n



        exception_thrown = 1
        try:
            await ircdns.start_n(1)
        except:
            exception_thrown = 1

        assert(exception_thrown)

        await ircdns.close()



        # Test partial start-continue
        ircdns = await IRCDNS(
            i=interface,
            seed=b"2" + IRC_SEED,
            clsSess=MockIRCSession,
            clsChan=MockIRCChan,
            servers=servers,
            executor=executor,
            do_shuffle=False
        ).start()

        await ircdns.start_n(2)
        assert(ircdns.p_sessions_next == 2)



        # Test making irc chan name
        dns_name = "p2pd_test"
        dns_tld = "test_tld"
        dns_val = "test val"
        await ircdns.pre_cache(dns_name, dns_tld)

        dns_hash = await ircdns.sessions[0].get_irc_chan_name(
            name=dns_name,
            tld=dns_tld,
            executor=executor
        )

        assert(irc_is_valid_chan_name(dns_hash))
        assert(len(dns_hash) <= 32)

        assert(await ircdns.get_server_len() == len(servers))
        assert(await ircdns.get_register_failure_max() == 2)

        assert(await ircdns.get_register_success_min() == 3)


        assert(await ircdns.get_register_success_max() == len(servers))


        assert(await ircdns.get_lookup_success_min() == 3)


        # Register, store, then get.
        ret, _ = await ircdns.name_register(dns_name, dns_tld)
        assert(len(ret))


        # Test store.
        await ircdns.store_value(dns_val, dns_name, dns_tld)
        otps = []
        sigs = []
        for i in range(0, len(servers)):
            test_hash = await ircdns.sessions[i].get_irc_chan_name(
                name=dns_name,
                tld=dns_tld
            )

            assert(test_hash in ircdns.sessions[i].chans)

            # Msg integrity is uniquely tied to channel integrity.
            topic_val = ircdns.sessions[i].chans[test_hash].pending_topic
            out = f_unpack_topic(
                test_hash,
                topic_val,
                ircdns.sessions[i]
            )

            # As a basic assumption these should all be different.
            # Time discards non-seconds so it may occur in the same 'tick'
            assert(out["time"] > 315360000)
            assert(dns_val in out["msg"])
            assert(out["otp"] not in otps)
            assert(out["sig"] not in sigs)
            otps.append(out["otp"])
            sigs.append(out["sig"])

        # Get results list.
        results, _ = await ircdns.n_name_lookups(
            await ircdns.get_lookup_success_min(),
            0,
            dns_name,
            dns_tld
        )


        best = await ircdns.n_more_or_best(results)

        # Check that best value is correct.
        highest = results[0]["time"]
        for r in results:
            if r["time"] > highest:
                highest = r["time"]

        assert(highest == best["time"])

        freshest = await ircdns.name_lookup(
            name=dns_name,
            tld=dns_tld
        )

        assert(best == freshest)

        await ircdns.close()

        # unpack

        #assert(ircdns.sessions[0].chans[dns_hash].pending_topic == dns_val)

        # assert len of user pass nick email ... etc

    async def test_with_some_failed_sessions(self):
        class MockIRCSession2(MockIRCSession):
            async def start(self, i):
                if self.db is not None:
                    last_started_key = self.db_key("last_started")
                    await self.db.put(last_started_key, time.time())


                if self.offset not in [1, 5]:
                    self.started.set_result(True)
                else:
                    raise Exception("Cannot start!")

        interface = None
        ircdns = await IRCDNS(
            i=interface,
            seed=b"3" + IRC_SEED,
            clsSess=MockIRCSession2,
            clsChan=MockIRCChan,
            servers=IRC_TEST_SERVERS_SEVEN,
            executor=executor,
            do_shuffle=False
        ).start()


        assert(await ircdns.get_server_len() == 7)
        assert(await ircdns.get_register_success_min() == 5)
        dns_name = "p2pd_test2"
        dns_tld = "test_tld2"
        dns_val = "test val2"
        ret, _ = await ircdns.name_register(dns_name, dns_tld)
        await ircdns.store_value(dns_val, dns_name, dns_tld)

        ret = await ircdns.name_lookup(dns_name, dns_tld)
        assert(dns_val in ret["msg"])
        await ircdns.close()

    async def test_register_partial_to_simulate_session_restart(self):
        """
        It should initially start enough sessions to pass the
        get_success_min check (p_session_next will point to len)
        but later when it calls start again it
        should return success that time indicating that servers have
        come back online. The results should show success for those previously
        down servers.
        """
        class MockIRCSession4(MockIRCSession):
            offset_count = {}
            async def start(self, i):
                if self.db is not None:
                    last_started_key = self.db_key("last_started")
                    await self.db.put(last_started_key, time.time())


                if self.offset not in self.offset_count:
                    self.offset_count[self.offset] = 0
                self.offset_count[self.offset] += 1
                

                # Simulate one server down.
                if self.offset not in [0, 1]:
                    self.started.set_result(True)
                else:
                    if self.offset_count[self.offset] > 1:
                        self.started.set_result(True)
                    else:
                        raise Exception("Start failure!")

            # Disable DB caching for this test.
            async def db_is_name_registered(self, name, tld, pw=""):
                return False

        interface = None
        ircdns = await IRCDNS(
            i=interface,
            seed=b"4" + IRC_SEED,
            clsSess=MockIRCSession4,
            clsChan=MockIRCChan,
            servers=IRC_TEST_SERVERS_SEVEN,
            executor=executor,
            do_shuffle=False
        ).start()

        await ircdns.start_n(len(IRC_TEST_SERVERS_SEVEN))
        assert(ircdns.p_sessions_next == len(IRC_TEST_SERVERS_SEVEN))
        ret, _ = await ircdns.name_register("test_name6", "test_tld6")
        for chan_info in ret:
            assert(chan_info["status"] == IRC_REGISTER_SUCCESS)

        await ircdns.close()

    async def test_successive_register(self):
        """
        The first register should not succeed on all servers (some will be down)
        while the second register should succeed to simulate the servers
        returning. One of the names will be unavailable just to mix things up.
        """
        class MockIRCSession5(MockIRCSession):
            offset_count = {}
            async def start(self, i):
                if self.db is not None:
                    last_started_key = self.db_key("last_started")
                    await self.db.put(last_started_key, time.time())


                if self.offset not in self.offset_count:
                    self.offset_count[self.offset] = 0
                self.offset_count[self.offset] += 1
                

                # Simulate one server down.
                if self.offset not in [0]:
                    self.started.set_result(True)
                else:
                    # Fail on start_n and start in do_register.
                    # Next register will succeed.
                    if self.offset_count[self.offset] > 2:
                        self.started.set_result(True)
                    else:
                        raise Exception("Start failure!")
                    
            # ... and simulate one server having a name conflict.
            async def is_chan_registered(self, chan_name):
                if self.offset in [1]:
                    return "someone_else"
                else:
                    return await super().is_chan_registered(chan_name)
                
            # Disable DB caching for this test.
            async def db_is_name_registered(self, name, tld, pw=""):
                return False
                
        interface = None
        ircdns = await IRCDNS(
            i=interface,
            seed=b"5" + IRC_SEED,
            clsSess=MockIRCSession5,
            clsChan=MockIRCChan,
            servers=IRC_TEST_SERVERS_SEVEN,
            executor=executor,
            do_shuffle=False
        ).start()

        dns_name = "dns_name8"; dns_tld = "dns_tld8"
        ret, _ = await ircdns.name_register(dns_name, dns_tld)
        assert(ret[0]["status"] == IRC_START_FAILURE)
        assert(ret[1]["status"] == IRC_REGISTER_FAILURE)
        assert(ret[2]["status"] == IRC_REGISTER_SUCCESS)

        ret, _ = await ircdns.name_register(dns_name, dns_tld)
        assert(ret[0]["status"] == IRC_REGISTER_SUCCESS)
        assert(ret[1]["status"] == IRC_REGISTER_FAILURE)
        assert(ret[2]["status"] == IRC_REGISTER_SUCCESS)
        await ircdns.close()

    # check partial availability of names (some are already taken)
    async def test_partial_availability_of_names(self):
        class MockIRCSession3(MockIRCSession):
            async def start(self, i):
                if self.db is not None:
                    last_started_key = self.db_key("last_started")
                    await self.db.put(last_started_key, time.time())


                # Simulate one server down.
                if self.offset not in [5]:
                    self.started.set_result(True)
                else:
                    raise Exception("Start failure!")

            # ... and simulate one server having a name conflict.
            async def is_chan_registered(self, chan_name):
                if self.offset in [2]:
                    return "someone_else"
                else:
                    return await super().is_chan_registered(chan_name)
                
            # Disable DB caching for this test.
            async def db_is_name_registered(self, name, tld, pw=""):
                return False

        interface = None
        ircdns = await IRCDNS(
            i=interface,
            seed=b"6" + IRC_SEED,
            clsSess=MockIRCSession3,
            clsChan=MockIRCChan,
            servers=IRC_TEST_SERVERS_SEVEN,
            executor=executor,
            do_shuffle=False
        ).start()

        dns_name = "p2pd_test3"
        dns_tld = "test_tld3"
        dns_val = "test val3"
        ret, _ = await ircdns.name_register(dns_name, dns_tld)
        assert(ret[2]["status"] == IRC_REGISTER_FAILURE)
        assert(ret[5]["status"] == IRC_START_FAILURE)
        assert(ret[0]["status"] == IRC_REGISTER_SUCCESS)
        await ircdns.store_value("initial val", dns_name, dns_tld)
        await ircdns.store_value(dns_val, dns_name, dns_tld)
        ret = await ircdns.name_lookup(dns_name, dns_tld)
        assert(dns_val in ret["msg"])
        await ircdns.close()

    async def test_to_check_db_integrity_multi_chan_reg(self):
        interface = None
        ircdns = await IRCDNS(
            i=interface,
            seed=b"7" + IRC_SEED,
            clsSess=MockIRCSession,
            clsChan=MockIRCChan,
            servers=IRC_TEST_SERVERS_SEVEN,
            executor=executor,
            do_shuffle=False
        ).start()

        ret, _ = await ircdns.name_register("test_name4", "tld4")
        ret, _ = await ircdns.name_register("test_name5", "tld5")

        # Get list of chans stored in first session.
        chan_list = await ircdns.sessions[0].db_load_chan_list()
        vectors = [
            ["test_name4", "tld4"],
            ["test_name5", "tld5"]
        ]

        # Should have found both dns names.
        for vector in vectors:
            dns = {
                "name": vector[0],
                "tld": vector[1],
                "pw": ""
            }

            found = False
            for chan_name in chan_list:
                chan_info_key = ircdns.sessions[0].db_key(f"chan/{chan_name}")
                chan_info = await ircdns.db.get(chan_info_key, {})
                if "dns" not in chan_info:
                    continue

                if chan_info["dns"] == dns:
                    found = True
                    break
            assert(found)


        await ircdns.close()

    async def test_irc_refresher(self):
        """
        We'll artificially set a particular chan to have a refresh time far in
        the past to force the refresh code to fire. The code that is fired to
        achieve refreshes will be monkey-patched to avoid network errors.
        Fixed structures will be stored in the DB to make the testing simpler.
        The refresh code is fairly straight forwards so easier to test.
        """

        ircdns = await IRCDNS(
            i=None,
            seed=b"8" + IRC_SEED,
            clsSess=MockIRCSession,
            clsChan=MockIRCChan,
            servers=IRC_TEST_SERVERS_SEVEN,
            executor=executor,
            do_shuffle=False
        ).start()
        await ircdns.start_n(len(IRC_TEST_SERVERS_SEVEN))

        chan_name = "expired_chan"
        chan_info = {
            "chan_name": chan_name,
            "last_refresh": time.time() - (365 * 24 * 60 * 60)
        }

        nick = "expired_nick"
        nick_info = {
            "nick": nick,
            "last_refresh": time.time() - (365 * 24 * 60 * 60)
        }

        db = ircdns.db

        # Store the above expired chan info in the db.
        for n in range(0, ircdns.p_sessions_next):
            s = ircdns.sessions[n]
            s.refresh_chan_fired = None
            s.refresh_nick_fired = None
            chan_list = await s.db_load_chan_list()
            chan_list.add(chan_name)
            
            # Simulate expired channel.
            key_name = s.db_key("chan_list")
            await db.put(key_name, chan_list)
            key_name = s.db_key(f"chan/{chan_name}")
            await db.put(key_name, chan_info)

            # Simulate expired nick.
            key_name = s.db_key("nick")
            await db.put(key_name, nick_info)

        # Skeleton functions just to know if they were triggered.
        async def refresh_chan(chan_name, session):
            session.refresh_chan_fired = chan_name
        async def refresh_nick(session):
            session.refresh_nick_fired = "1"

        # Patch refresher obj with skeletons above.
        refresher = IRCRefresher(ircdns)
        refresher.refresh_chan = refresh_chan
        refresher.refresh_nick = refresh_nick

        # Call refresher for first time.
        # It should run the skeleton refresh funcs.
        await refresher.refresher()

        # Check that refresh functions fired.
        for n in range(0, ircdns.p_sessions_next):
            s = ircdns.sessions[n]
            assert(s.refresh_chan_fired == chan_name)
            assert(s.refresh_nick_fired == "1")
            s.refresh_chan_fired = None
            s.refresh_nick_fired = None

        # Run refresh again.
        # It should do nothing this time.
        await refresher.refresher()

        # Now check refresh function WASNT fired.
        for n in range(0, ircdns.p_sessions_next):
            s = ircdns.sessions[n]
            assert(s.refresh_chan_fired == None)
            assert(s.refresh_nick_fired == None)

        await ircdns.close()
    
    # python -m unittest test_irc_dns.TestIRCDNS.test_irc_refresher_does_register
    async def test_irc_refresher_does_register(self):
        class MockIRCSession6(MockIRCSession):
            offset_count = {}
            async def start(self, i):
                if self.db is not None:
                    last_started_key = self.db_key("last_started")
                    await self.db.put(last_started_key, time.time())


                if self.offset not in self.offset_count:
                    self.offset_count[self.offset] = 0
                self.offset_count[self.offset] += 1
                

                # Simulate one server down.
                if self.offset not in [0]:
                    self.started.set_result(True)
                else:
                    # Fail on start_n and start in do_register.
                    # Next register will succeed.
                    if self.offset_count[self.offset] > 2:
                        self.started.set_result(True)
                    else:
                        raise Exception("Start failure!")
                
            # Disable DB caching for this test.
            async def db_is_name_registered(self, name, tld, pw=""):
                return False
            
        ircdns = await IRCDNS(
            i=None,
            seed=b"9" + IRC_SEED,
            clsSess=MockIRCSession6,
            clsChan=MockIRCChan,
            servers=IRC_TEST_SERVERS_SEVEN,
            executor=executor,
            do_shuffle=False
        ).start()

        dns_name = "dns_name9"; dns_tld = "dns_tld9"
        ret, _ = await ircdns.name_register(dns_name, dns_tld)
        assert(ret[0]["status"] == IRC_START_FAILURE)
        assert(ret[1]["status"] == IRC_REGISTER_SUCCESS)
        dns = ret[0]["dns"]

        refresher = IRCRefresher(ircdns)
        refresher.register_chan = lambda x: None
        await refresher.refresher(ignore_failure=True)

        # Check that register was ran for the channel in refresh.
        register_success = False
        chan_list = await ircdns.sessions[0].db_load_chan_list()
        for chan_name in chan_list:
            chan_info_key = ircdns.sessions[0].db_key(f"chan/{chan_name}")
            chan_info = await ircdns.db.get(chan_info_key, {})
            if "dns" not in chan_info:
                continue

            if chan_info["dns"] == dns:
                if chan_info["status"] == IRC_REGISTER_SUCCESS:
                    register_success = True
        assert(register_success)

        await ircdns.close()

    async def test_nick_details_save(self):
        ircdns = await IRCDNS(
            i=None,
            seed=b"10" + IRC_SEED,
            clsSess=MockIRCSession,
            clsChan=MockIRCChan,
            servers=IRC_TEST_SERVERS_SEVEN,
            executor=executor,
            do_shuffle=False
        ).start()

        # Start only a single session at offset 0.
        await ircdns.start_n(1)

        # Attempt to fetch sessions nick details.
        nick_key = ircdns.sessions[0].db_key("nick")
        nick_info = await ircdns.db.get(nick_key, {})
        vectors = ["domain", "nick", "username", "user_pass"]
        vectors += ["email", "last_refresh"]
        for vector in vectors:
            assert(vector in nick_info)

        # Cleanup.
        await ircdns.close()

    async def test_irc_dns_real_world(self):
        return
        interface = await Interface()
        seed = b_sha3_256(b"some secret seed")
        ircdns = await IRCDNS(
            i=interface,
            seed=seed,
            servers=IRC_SERVERS
        ).start()
        refresher = IRCRefresher(ircdns)

        # Register a new name from scratch.
        name_info = ["muh awesome name", "node"]
        await ircdns.name_register(*name_info)

        # Do two different write/reads.
        for _ in range(0, 2):
            # Store value in entry.
            name_val  = rand_plain(10)
            await ircdns.store_value(*[name_val] + name_info)

            # Lookup its value and check results.
            ret = await ircdns.name_lookup(*name_info)
            assert(to_b(ret["msg"]) == to_b(name_val))

        # Test refresher also doesn't crash.
        await refresher.refresher()

        # Cleanup the dns manager.
        await ircdns.close()

    async def test_irc_servers_work(self):
        return
        dns_value = "Test dns val."
        dns_name = "testing_dns_name"
        dns_tld = "testing"
        dns_pw = ""
        executor = ProcessPoolExecutor()
        server_list = IRC_SERVERS

        seed = b"test_seed2" * 20
        i = await Interface().start()
        ircdns = await IRCDNS(
            i,
            seed,
            server_list,
            executor,
            do_shuffle=False
        ).start()

        


        print("If start")
        print(i)

        for offset, s in enumerate(server_list[3:]):
            print(f"testing {s} : {offset}")

            ses = IRCSession(s, seed)

            try:
                await ses.start(i)
                print("start success")
            except:
                print(f"start failed for {s}")
                what_exception()


            chan_name = await ses.get_irc_chan_name(
                name=dns_name,
                tld=dns_tld,
                pw=dns_pw,
                executor=executor
            )

            # Test chan create.
            print("trying to check if chan is registered.")
            ret = await ses.is_chan_registered(chan_name)
            if ret:
                print(f"{chan_name} registered, not registering")

                # Check channel owner returned is correct.
                assert(ret == ses.nick)

                # 'load' chan instead.
                irc_chan = IRCChan(chan_name, ses)
                ses.chans[chan_name] = irc_chan
                print(irc_chan.chan_pass)
            else:
                print(f"{chan_name} not registered, attempting to...")
                await ses.register_chan(chan_name)
                ret = await ses.is_chan_registered(chan_name)
                if ret:
                    print("success")
                else:
                    print("failure")
                    exit()

            # Test set topic.
            chan_topic = to_s(rand_plain(8))
            chan_topic, _ = await f_pack_topic(
                value=dns_value,
                name=dns_name,
                tld=dns_tld,
                pw=dns_pw,
                ses=ses,
                clsChan=IRCChan,
                executor=executor
            )

            print(f"trying to store {chan_topic}")

            #await irc_dns.chans[chan_name].get_ops()
            #print("get ops done")
            await ses.chans[chan_name].set_topic(chan_topic)
            print("set topic done")

            # Potential race condition between getting new chan.
            await asyncio.sleep(4)

            outside_user = IRCSession(s, seed + "2")
            try:
                await outside_user.start(i)
                print("start success")
            except:
                print(f"start failed for outside user")
                what_exception()

            out = await outside_user.get_chan_topic(chan_name)
            if out != chan_topic:
                print(f"got {out} for chan topic and not {chan_topic}")
                exit()
            else:
                print("success")

            # Test decode of encoded chan topic.
            unpack_topic = f_unpack_topic(
                chan_name,
                chan_topic,
                ses
            )

            print(unpack_topic)

            print("unpacked topic = ")
            assert(to_b(unpack_topic["msg"]) == to_b(dns_value))

            # Cleanup.
            await outside_user.close()
            await ses.close()
            input("Press enter to test next server.")
            input()

    async def test_find_more_servers(self):
        pass

if __name__ == '__main__':
    executor = ProcessPoolExecutor()
    main()
    executor.shutdown(wait=False)

"""

"""


from p2pd.test_init import *

from p2pd import *

class TestP2PAddr(unittest.IsolatedAsyncioTestCase):
    async def test_p2p_addr(self):
        x = b"0,1-[0,8.8.8.8,192.168.21.200,33334,2,5,1]-0-kjCimIvudifJh7X"
        out = parse_peer_addr(x)
        print(out)
        self.assertTrue(len(out[IP4]))

    async def test_packing(self):
        node_id = b"123" * 10
        signal_offsets = [1, 2, 3]
        if_list = await load_interfaces()

        print(if_list)

        addr_buf = pack_peer_addr(
            node_id,
            if_list,
            signal_offsets
        )

        out = unpack_peer_addr(addr_buf)
        print(out)

if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-
"""
basen
Originated from http://blog.suminb.com/archives/558
"""


__author__ = "Sumin Byeon"
__email__ = "suminb@gmail.com"
__version__ = "1.0.0"

# Used for IRC ident passwords.
B64_CHARSET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"

# Used for chan names.
B36_CHARSET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"

# Used for chan topics.
# Modified to remove : which is used for prefix and suffix in IRC.
B92_CHARSET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz~!@#$%^&*()_+{}\"<>?`-=[];',./|"

def encode(n, charset=B64_CHARSET):
    """Encodes a given integer ``n``."""

    chs = []
    while n > 0:
        n, r = divmod(n, len(charset))
        chs.insert(0, charset[r])

    if not chs:
        return "0"

    return "".join(chs)


def encodebytes(barray, charset=B64_CHARSET):
    """Encodes a bytestring into a base62 string.

    :param barray: A byte array
    :type barray: bytes
    :rtype: str
    """

    _check_type(barray, bytes)

    # Count the number of leading zeros.
    leading_zeros_count = 0
    for i in range(len(barray)):
        if barray[i] != 0:
            break
        leading_zeros_count += 1

    # Encode the leading zeros as "0" followed by a character indicating the count.
    # This pattern may occur several times if there are many leading zeros.
    n, r = divmod(leading_zeros_count, len(charset) - 1)
    zero_padding = f"0{charset[-1]}" * n
    if r:
        zero_padding += f"0{charset[r]}"

    # Special case: the input is empty, or is entirely null bytes.
    if leading_zeros_count == len(barray):
        return zero_padding

    value = encode(int.from_bytes(barray, "big"), charset=charset)
    return zero_padding + value


def decode(encoded, charset=B64_CHARSET):
    """Decodes a base62 encoded value ``encoded``.

    :type encoded: str
    :rtype: int
    """
    _check_type(encoded, str)

    base = len(charset)
    l, i, v = len(encoded), 0, 0
    for x in encoded:
        v += _value(x, charset=charset) * (base ** (l - (i + 1)))
        i += 1

    return v


def decodebytes(encoded, charset=B64_CHARSET):
    """Decodes a string of base62 data into a bytes object.

    :param encoded: A string to be decoded in base62
    :type encoded: str
    :rtype: bytes
    """

    leading_null_bytes = b""
    while encoded.startswith("0") and len(encoded) >= 2:
        leading_null_bytes += b"\x00" * _value(encoded[1], charset)
        encoded = encoded[2:]

    decoded = decode(encoded, charset=charset)
    buf = bytearray()
    while decoded > 0:
        buf.append(decoded & 0xFF)
        decoded //= 256
    buf.reverse()

    return leading_null_bytes + bytes(buf)


def _value(ch, charset):
    """Decodes an individual digit of a base encoded string."""

    try:
        return charset.index(ch)
    except ValueError:
        raise ValueError(f"base{len(charset)}: Invalid character (%s)" % ch)


def _check_type(value, expected_type):
    """Checks if the input is in an appropriate type."""

    if not isinstance(value, expected_type):
        msg = "Expected {} object, not {}".format(
            expected_type, value.__class__.__name__
        )
        raise TypeError(msg)


from p2pd import *

async def example():
    # Start the default interface.
    i = await Interface().start()
    #
    # Load additional NAT details.
    # Restrict, random port NAT assumed by default.
    await i.load_nat()
    #
    # Show the interface details.
    repr(i)

if __name__ == '__main__':
    async_test(example)


from p2pd import *

async def example():
    #
    # Start default interface.
    # Don't bother resolving external addresses.
    i = await Interface().start()
    #
    # Echo server address.
    route = await i.route().bind()
    echo_dest = await Address("tcpbin.com", 4242, route).res()
    #
    # Open a connection to the echo server.
    pipe = await pipe_open(TCP, route, echo_dest)
    # No need to call pipe.subscribe(SUB_ALL).
    # This is done automatically if a destination is provided.
    # Call pipe.unsubscribe(SUB_ALL) to turn this off.
    #
    # Send data down the pipe.
    msg = b"do echo test"
    await pipe.send(msg + b"\r\n", echo_dest.tup)
    #
    # Receive data back.
    data = await pipe.recv(SUB_ALL, 4)
    assert(msg in data)
    #
    # Close the sockets.
    await pipe.close()

# Utility function to run an async function.
if __name__ == '__main__':
    async_test(example)


import struct
from .settings import *
from .nat import *
from .ip_range import *


# No more than n interfaces per address family in peer addr.
PEER_ADDR_MAX_INTERFACES = 4

# No more than n signal pipes to send signals to nodes.
SIGNAL_PIPE_NO = 3

"""
        can be up to N interfaces
[ IP4 nics
    [
        interface_offset,
        ext ip,
        nic ip,
        port,
        nat_type,
        delta_type,
        delta_val
    ]
    ,... more interfaces for AF family
],[IP6 nics ...],node_id
"""
def make_peer_addr(node_id, interface_list, signal_offsets, port=NODE_PORT, ip=None, nat=None, if_index=None):
    ensure_resolved(interface_list)
    signal_offsets_as_str = [to_b(str((x))) for x in signal_offsets]
    bufs = [
        # Make signal pipe buf.
        b','.join(signal_offsets_as_str)
    ]

    for af in [IP4, IP6]:
        af_bufs = []
        for i, interface in enumerate(interface_list):
            # AF type is not supported.
            if not len(interface.rp[af].routes):
                continue

            r = interface.route(af)
            if r is None:
                continue

            if nat:
                nat_type = nat["type"]
                delta_type = nat["delta"]["type"]
                delta_value = nat["delta"]["value"]
            else:
                nat_type = interface.nat["type"]
                delta_type = interface.nat["delta"]["type"]
                delta_value = interface.nat["delta"]["value"]

            af_bufs.append(b"[%d,%b,%b,%d,%d,%d,%d]" % (
                if_index or i,
                ip or to_b(r.ext()),
                ip or to_b(r.nic()),
                port,
                nat_type,
                delta_type,
                delta_value
            ))

        if len(af_bufs):
            af_bufs = b'|'.join(af_bufs)
        else:
            af_bufs = b''
        
        # The as_buf may be empty if AF has no routes.
        # Expected and okay.
        bufs.append(af_bufs or b"0")
    
    bufs.append(node_id)
    return b'-'.join(bufs)

"""
New packed binary format for addresses. 
Doing this was necessary for the IRC DNS module as space in
topics for records is quite scarce. There's room here for improvement
if you wanted to make each bit count. But I haven't gone overboard here.
"""
def pack_peer_addr(node_id, interface_list, signal_offsets, port=NODE_PORT, ip=None, nat=None, if_index=None):
    # Truncate node id to 8.
    node_id = node_id[:8]

    """
    Indicates variable length sections (list of signal offsets
    and number of interfaces.) Encodes 1 byte integers as byte
    characters so they can be stored in an unsigned char.
    """
    args = [
        # Node ID(8), lport(2), signal offsets(1), if no(1)
        # signal offsets[] ...
        "HBB" + ("B" * len(signal_offsets)),
        port,
        len(signal_offsets),
        len(interface_list)
    ]
    for offset in signal_offsets:
        args.append(offset)
    
    # Pack header portion.
    buf = node_id
    buf += struct.pack(*args)

    # if_no, ip4, null or ..., ip6, null or ...
    for i, interface in enumerate(interface_list):
        # Append interface no.
        buf += bytes([if_index or i])

        # Loop over AFs.
        for af in [IP4, IP6]:
            # Append AF type.
            buf += bytes([af])

            # AF type is not supported.
            if not len(interface.rp[af].routes):
                buf += bytes([0])
                continue

            # AF type not supported.
            r = interface.route(af)
            if r is None:
                buf += bytes([0])
                continue

            # Indicate segment is filled.
            buf += bytes([1])

            # Main details for this interface.
            if nat:
                nat_type = nat["type"]
                delta_type = nat["delta"]["type"]
                delta_value = nat["delta"]["value"]
            else:
                nat_type = interface.nat["type"]
                delta_type = interface.nat["delta"]["type"]
                delta_value = interface.nat["delta"]["value"]

            # Pack interface details.
            # nat type 1 - 7
            # delta type 1 - 7
            # delta value +/- port
            buf += struct.pack(
                "BBl",
                nat_type,
                delta_type,
                delta_value
            )

            # Convert IPs to bytes.
            buf += bytes(IPRange(ip or r.nic()))
            buf += bytes(IPRange(ip or r.ext()))

    return buf

def validate_peer_addr(addr):
    # Check signal server offsets.
    for offset in addr["signal"]:
        if not in_range(offset, [0, len(MQTT_SERVERS) - 1]):
            log("p2p addr signal offset outside server range.")
            return None
        
    # Too many signal pipe offsets.
    if len(addr["signal"]) > SIGNAL_PIPE_NO:
        log("p2p addr invalid signal no for p2p addr")
        return None
    
    for af in VALID_AFS:
        for if_info in addr[af]:
            # Is listen port right?
            if not in_range(if_info["port"], [1, MAX_PORT]):
                log("p2p addr: listen port invalid")
                return None
            
            # Check NAT type is valid.
            if not in_range(if_info["nat"]["type"], [OPEN_INTERNET, BLOCKED_NAT]):
                log("p2p addr: nat type invalid")
                return None

            # Check delta type is valid.
            if not in_range(if_info["nat"]["delta"]["type"], [NA_DELTA, RANDOM_DELTA]):
                log("p2p addr: delta type invalid")
                return None

            # Check delta value is valid.
            delta = if_info["nat"]["delta"]["value"]
            delta = -delta if delta < 0 else delta
            if not in_range(delta, [0, MAX_PORT]):
                log("p2p addr: Delta value invalid")
                return None
            
    return addr

def unpack_peer_addr(addr):
    # Unpack header portion.
    node_id = addr[:8]; p = 8;
    port, signal_no, if_no = struct.unpack("HBB", addr[p:p + 4]); p += 4;

    # Unpack signal offsets (variable length.)
    signal_offsets = struct.unpack("B" * signal_no, addr[p:p + signal_no])
    out = {
        IP4: [],
        IP6: [],
        "node_id": node_id,
        "signal": signal_offsets
    }

    # Unpack if list.
    p += signal_no;
    for _ in range(0, if_no):
        if_index = addr[p]; p += 1;
        for _ in range(0, 2):
            # Get AF at pointer.
            af = addr[p]; p += 1

            # Address type for IF unsupported.
            if addr[p] == 0:
                # Next byte is another AF or if_index.
                p += 1
                continue
            else:
                # Segment is supported (and filled.)
                # Next byte starts the segment.
                p += 1

            # Unpack interface details.
            parts = struct.unpack("BBl", addr[p:p + 8]); p += 8;
            nat_type = parts[0]
            delta_type = parts[1]
            delta_value = parts[2]

            # Determine IP field sizes based on AF.
            if af == IP4:
                ip_size = 4
            if af == IP6:
                ip_size = 16

            # Exact IP portions.
            b_nic_ip = addr[p:p + ip_size]; p += ip_size;
            b_ext_ip = addr[p:p + ip_size]; p += ip_size;

            # Build dictionary of results.
            delta = delta_info(delta_type, delta_value)
            nat = nat_info(nat_type, delta)
            as_dict = {
                "if_index": if_index,
                "ext": IPRange(b_to_i(b_ext_ip)),
                "nic": IPRange(b_to_i(b_nic_ip)),
                "nat": nat,
                "port": port
            }

            # Save results.
            out[af].append(as_dict)

    return validate_peer_addr(out)

def parse_peer_addr(addr):
    af_parts = addr.split(b'-')
    if len(af_parts) != 4:
        log("p2p addr invalid parts")
        return None

    # Parse signal server offsets.
    p = af_parts.pop(0).split(b",")
    signal = [int(n) for n in p if in_range(int(n), [0, len(MQTT_SERVERS) - 1])]

    # Parsed dict.
    schema = [is_no, is_b, is_b, is_no,  is_no, is_no, is_no]
    translate = [to_n, to_b, to_b, to_n, to_n, to_n, to_n]
    out = {
        IP4: [],
        IP6: [],
        "node_id": af_parts[2],
        "signal": signal
    }

    for af_index, af_part in enumerate(af_parts[:2]):
        interface_infos = af_part.split(b'|')
        for info in interface_infos:
            # Strip outer braces.
            if len(info) < 2:
                continue
            inner = info[1:-1]

            # Split into components.
            parts = inner.split(b',')
            if len(parts) != 7:
                log("p2p addr: invalid parts no.")
                continue

            # Test type of field.
            # Convert to its end value if it passes.
            try:
                for j, part in enumerate(parts):
                    if not schema[j](part):
                        raise Exception("Invalid type.")
                    else:
                        parts[j] = translate[j](part)
            except Exception:
                continue

            # Is it a valid IP?
            try:
                IPRange(to_s(parts[1]))
            except Exception:
                log("p2p addr: ip invalid.")
                continue

            # Is it a valid IP?
            try:
                IPRange(to_s(parts[2]))
            except Exception:
                log("p2p addr: ip invalid.")
                continue
                    
            # Build dictionary of results.
            delta = delta_info(parts[5], parts[6])
            nat = nat_info(parts[4], delta)
            as_dict = {
                "if_index": parts[0],
                "ext": IPRange(to_s(parts[1])),
                "nic": IPRange(to_s(parts[2])),
                "nat": nat,
                "port": parts[3]
            }

            # Save results.
            af = VALID_AFS[af_index]
            out[af].append(as_dict)

    return validate_peer_addr(out)

def peer_addr_extract_exts(p2p_addr):
    exts = []
    for af in VALID_AFS:
        for info in p2p_addr[af]:
            exts.append(info["ext"])
            exts.append(info["nic"])

    return exts

def is_p2p_addr_us(addr_bytes, if_list):
    # Parse address bytes to address.
    addr = parse_peer_addr(addr_bytes)

    # Check all address families.
    for af in VALID_AFS:
        # Check all interface details for AF.
        for info in addr[af]:
            # Compare the external address.
            ipr = info["ext"]

            # Set the right interface to check.
            if_index = info["if_index"]
            if if_index + 1 > len(if_list):
                continue

            # Check all routes in the interface.
            interface = if_list[if_index]
            for route in interface.rp[af].routes:
                # Only interested in the external address.
                for ext_ipr in route.ext_ips:
                    # IPs are equal or in same block.
                    if ipr in ext_ipr:
                        return True

    # Nothing found that matches.
    return False

if __name__ == "__main__": # pragma: no cover
    from .interface import Interface
    async def test_p2p_addr():
        x = await Interface("enp3s0").start()
        if_list = [x]
        node_id = b"noasdfosdfo"
        b_addr = make_peer_addr(node_id, if_list)
        print(b_addr)

        addr = parse_peer_addr(b_addr)
        print(addr)

    async_test(test_p2p_addr)



from p2pd.test_init import *
from p2pd.net import *
from p2pd.interface import *
from p2pd.base_stream import *
from p2pd.rest_api import *
from p2pd.http_client_lib import *

class TestP2PDServer(unittest.IsolatedAsyncioTestCase):
    async def test_p2pd_server(self):
        # Start the P2PD server.
        i = await Interface().start()
        r = await i.route().bind()
        server = await start_p2pd_server([i], r, port=P2PD_PORT, do_loop=False, do_init=False, enable_upnp=False)

        # Make server implement a custom ping protocol extension.
        async def proto_extension(msg, client_tup, pipe):
            if b"PING" in msg:
                await pipe.send(b"PONG")
        server.node.add_msg_cb(proto_extension)

        # Server address.
        dest = await Address(r.bind_ip(), P2PD_PORT, r).res()

        # List of API end points to check.
        en = lambda x: to_s(urlencode(x))
        sub = en("[hH]e[l]+o")
        c = "con_name"
        urls = [
            # Some stat stuff.
            "/version",
            "/ifs",

            # Open new con.
            "/p2p/open/" + c + "/self",
            "/p2p/con/" + c,

            # Test basic text API stuff.
            "/p2p/sub/" + c + "/msg_p/" + sub,
            "/p2p/send/" + c + "/" + en("ECHO Hello, world!"),
            "/p2p/recv/" + c + "/msg_p/" + sub + "/timeout/2",

            # Test connection close works.
            "/p2p/close/" + c
        ]

        # Just load a bunch of URLs and check for errors.
        # Not the best unit test but over it.
        for url in urls:
            # Convert HTTP response to JSON.
            r = i.route()
            _, resp = await http_req(
                route=r, dest=dest, path=url,
                do_close=True
            )
            out = resp.out()
            j = json.loads(out)

            # IE: no error set.
            assert(j["error"] == 0)

        # Make a new con.
        c2 = "pipe_test"
        r = i.route()
        await http_req(r, dest, "/p2p/open/" + c2 + "/self", do_close=True)

        # Request a new pipe to a named p2p con.
        # This is a cool feature.
        r = i.route()
        p, out = await http_req(r, dest, "/p2p/pipe/" + c2, do_close=0)
        msg = b"this is a test"
        await p.send(b"ECHO " + msg)
        got = await p.recv(SUB_ALL, timeout=3)
        assert(msg in got)

        # Test custom protocol extension works.
        await p.send(b"PING")
        out = await p.recv()
        assert(b"PONG" in out)
        await p.close()

        # TODO: Test binary stuff.

        # Cleanup.
        await server.close()


if __name__ == '__main__':
    main()

"""
To do the pipe tests i need to be able to
send http requests on my own pipes
"""


import platform
from p2pd.test_init import *
from p2pd.utils import *
from p2pd.net import VALID_AFS
from p2pd.win_netifaces import *

from toxiclient import *
from toxiserver import *

class ToxicSlicer(ToxicBase):
    def set_params(self, avg_size, size_var, delay):
        self.avg_size = avg_size
        self.size_var = size_var
        self.delay = delay
        return self
    
    def chunk_old(self, start, end):
        """
        Base case:
        If the size is within the random variation,
        or already less than the average size, just
        return it. Otherwise split the chunk in about
        two, and recurse.
        """
        seg_len = end - start
        if seg_len - self.avg_size <= self.size_var:
            return [start, end]

        mid = int(start + (seg_len / 2))
        if self.size_var > 0:
            rand_len = rand_rang(0, (self.size_var * 2) + 1)
            mid += rand_len - self.size_var
            #mid = min(max(mid, start), end)

        # Recursion limit.
        left = self.chunk(start, mid)
        right = self.chunk(mid, end)

        return left + right
    
    def chunk(self, start, end):
        """
        Base case:
        If the size is within the random variation,
        or already less than the average size, just
        return it. Otherwise split the message into
        chunks of average size +/- size variation.
        """
        offsets = []
        seg_len = end - start
        if seg_len - self.avg_size <= self.size_var:
            return [start, end]
        
        # Build chunk offset list.
        # End offset overlaps with start offset.
        # This is not a mistake.
        p_start = start
        no = int(seg_len / self.avg_size)
        for _ in range(no):
            # Calculate a random variation.
            # May be positive or negative.
            rand_len = rand_rang(0, (self.size_var * 2) + 1)
            change = rand_len - self.size_var

            # Increase start pointer by random variation.
            p_end = p_start + (self.avg_size + change)

            # Record the chunk offsets.
            offsets += [p_start, min(p_end, end)]

            # Reached the end of the message size.
            if offsets[-1] == end:
                return offsets

            # Increase pointer for next segment.
            p_start = offsets[-1]

        return offsets

    async def run(self, msg, dest_pipe):
        offsets = self.chunk(0, len(msg))
        if self.delay:
            await asyncio.sleep(self.delay / 1000)

        for i in range(1, len(offsets), 2):
            chunk = msg[offsets[i-1]:offsets[i]]
            if self.delay:
                await asyncio.sleep(self.delay / 1000)

            #await dest_pipe.send(chunk)
            print(chunk)

class TestSlicer(unittest.IsolatedAsyncioTestCase):
    async def test_slicer(self):
        print("in test slicer.")
        msg = b"this is a sample message to slice."
        slicer = ToxicSlicer().set_params(
            avg_size=4,
            size_var=2,
            delay=10
        )

        await slicer.run(msg, None)




if __name__ == '__main__':
    main()



import asyncio
import copy
from .p2p_addr import *
from .tcp_punch import *
from .turn_client import *
from .signaling import *

"""
TCP 
nc -4 -l 127.0.0.1 10001 -k
nc -6 -l ::1 10001 -k
"""

P2P_DIRECT = 1
P2P_REVERSE = 2
P2P_PUNCH = 3
P2P_RELAY = 4

# TURN is not included as a default strategy because it uses UDP.
# It will need a special explanation for the developer.
# SOCKS might be a better protocol for relaying in the future.
P2P_STRATEGIES = [P2P_DIRECT, P2P_REVERSE, P2P_PUNCH]

async def get_tcp_hole(side, pipe_id, node_id, punch_client, node):
    pipe = None
    try:
        pipe = await punch_client.proto_do_punching(
            side,
            node_id,
            pipe_id,
            node.msg_cb
        )
    except:
        log_exception()

    if pipe is not None:
        # Save pipe reference.
        node.pipes[pipe_id] = pipe

        # Notify any waiters.
        if pipe_id in node.pipe_events:
            node.pipe_events[pipe_id].set()

        # Add cleanup callback.
        pipe.add_end_cb(node.rm_pipe_id(pipe_id))

    return pipe

"""
If nodes are behind the same router they will have
the same external address. Using this address for
connections will fail because it will be the same
address as ourself. The solution here is to replace
that external address with a private, NIC address.
For this reason the P2P address format includes
a private address section that corrosponds to
the address passed to bind() for the nodes listen().
"""
def work_behind_same_router(src_addr, dest_addr):
    new_addr = copy.deepcopy(dest_addr)
    for af in VALID_AFS:
        for dest_info in new_addr[af]:
            for src_info in src_addr[af]:
                # Same external address as one of our own.
                if dest_info["ext"] == src_info["ext"]:
                    # Connect to its internal address instead.
                    dest_info["ext"] = dest_info["nic"]

                    # Disable NAT in LANs.
                    delta = delta_info(NA_DELTA, 0)
                    nat = nat_info(OPEN_INTERNET, delta)
                    dest_info["nat"] = nat

    return new_addr

def build_reverse_msg(pipe_id, addr_bytes, b_proto=b"TCP"):
    out = b"P2P_DIRECT %s %s %s" % (pipe_id, b_proto, addr_bytes) 
    return out

def build_punch_response(cmd, pipe_id, punch_ret, src_addr_bytes, af, our_if_index, their_if_index):
    def mappings_to_bytes(mappings):
        pairs = []
        for pair in mappings:
            # remote, reply, local.
            pairs.append( b"%d,%d,%d" % (pair[0], pair[1], pair[2]) )

        return b"|".join(pairs)

    #      cm pip se nt pr ad af (if indexs)
    out = b"%s %s %d %s %s %s %d %d %d" % (
        cmd,
        pipe_id,
        0, # session id.
        to_b(str(punch_ret[1])), # ntp
        mappings_to_bytes(punch_ret[0]), # predictions
        src_addr_bytes,
        af,
        our_if_index,
        their_if_index
    )

    return out

def parse_prediction_str(s_predictions):
    predictions = []
    prediction_strs = s_predictions.split("|")
    for prediction_str in prediction_strs:
        remote_s, reply_s, local_s = prediction_str.split(",")
        prediction = [to_n(remote_s), to_n(reply_s), to_n(local_s)]
        if not in_range(prediction[0], [1, MAX_PORT]):
            raise Exception(f"Invalid remote port {prediction[0]}")

        if not in_range(prediction[-1], [1, MAX_PORT]):
            raise Exception(f"Invalid remote port {prediction[-1]}")

        predictions.append(prediction)

    if not len(predictions):
        raise Exception("No predictions received.")

    return predictions

def parse_punch_response(parts):
    return {
        "cmd": parts[0],
        "pipe_id": to_b(parts[1]),
        "session_id": to_n(parts[2]),
        "ntp_time": Dec(parts[3]),
        "predictions": parse_prediction_str(parts[4]),
        "src_addr": parse_peer_addr(to_b(parts[5])),
        "src_addr_bytes": to_b(parts[5]),
        "af": to_n(parts[6]),
        "if": {
            "them": to_n(parts[7]),
            "us": to_n(parts[8])
        }
    }

f_punch_index = lambda ses, addr, pipe: b"%d %b %b" % (ses, addr, pipe)
class P2PPipe():
    def __init__(self, node):
        self.node = node
        self.tasks = []

    # Minor data structure cleanup.
    def cleanup_pipe(self, pipe_id):
        def closure():
            if pipe_id in self.node.pipes:
                del self.node.pipes[pipe_id]

        return closure

    # Waits for a pipe event to fire.
    # Returns the corrosponding pipe.
    async def pipe_waiter(self, pipe_id, timeout=0):
        try:
            pipe_event = self.node.pipe_events[pipe_id]
            if timeout:
                await asyncio.wait_for(
                    pipe_event.wait(),
                    timeout
                )
            else:
                await pipe_event.wait()
            pipe = self.node.pipes[pipe_id]
            return pipe
        except asyncio.TimeoutError:
            log(f"> Pipe waiter timeout {pipe_id} {timeout}")
            return None
        
    # Implements the p2p connection techniques.
    async def open_pipe(self, pipe_id, dest_bytes, strategies=P2P_STRATEGIES):
        log(f"> Pipe open = {pipe_id}; {dest_bytes}; {strategies}")

        # Setup events for waiting on the pipe.
        pipe = None
        p2p_dest = parse_peer_addr(dest_bytes)
        self.node.pipe_events[pipe_id] = asyncio.Event()

        # See if a matching signal pipe exists.
        signal_pipe = self.node.find_signal_pipe(p2p_dest)
        if signal_pipe is None:
            for offset in p2p_dest["signal"]:
                # Build a channel used to relay signal messages to peer.
                mqtt_server = MQTT_SERVERS[offset]
                signal_pipe = SignalMock(
                    peer_id=to_s(self.node.node_id),
                    f_proto=self.node.signal_protocol,
                    mqtt_server=mqtt_server
                )

                # If it fails unset the client.
                try:
                    # If it's successful exit server offset attempts.
                    await signal_pipe.start()
                    self.node.signal_pipes[offset] = signal_pipe
                    break
                except asyncio.TimeoutError:
                    # Cleanup and make sure it's unset.
                    await signal_pipe.close()
                    signal_pipe = None

        # Check if a signal pipe was found.
        if signal_pipe is None:
            raise Exception("Unable to open same signal pipe as peer.")

        # Patch dest address if it's behind same router.
        dest_patch = work_behind_same_router(self.node.p2p_addr, p2p_dest)
        timeout = 0
        for strategy in strategies:
            # Try connect directly to peer.
            if strategy == P2P_DIRECT:
                timeout = 4
                open_task = self.direct_connect(dest_patch, pipe_id)

            # Tell the peer to try connect to us.
            if strategy == P2P_REVERSE:
                # Associate the pending pipe with external IP(s).
                exts = peer_addr_extract_exts(p2p_dest)
                self.node.expected_addrs[pipe_id] = exts

                # Instruct the peer to do the reverse connect.
                out = build_reverse_msg(pipe_id, self.node.addr_bytes)
                await signal_pipe.send_msg(
                    out,
                    to_s(p2p_dest["node_id"])
                )

                # Wait for the new pipe to appear.
                timeout = 10
                open_task = self.pipe_waiter(pipe_id)

            # Do TCP hole punching to try connect.
            if strategy == P2P_PUNCH:
                # Start TCP hole punching.
                self.tasks.append(
                    asyncio.create_task(
                        async_wrap_errors(
                            self.tcp_hole_punch(dest_patch, pipe_id, signal_pipe)
                        )
                    )
                )

                # Wait for the hole to appear.
                """
                Meet delay = 3
                Punch duration = 3
                Update wait = 4
                Init wait = 4
                Padding = 6
                """
                timeout = 30
                open_task = self.pipe_waiter(pipe_id)

            # Use a TURN server to connect to a peer.
            if strategy == P2P_RELAY:
                # 3 retries, up to 10 secs for each attempt.
                timeout = 45
                open_task = self.udp_relay(p2p_dest, pipe_id, signal_pipe)

            # Got valid pipe so return it.
            pipe = await async_wrap_errors(open_task, timeout)
            if pipe is not None:
                log(f"> Got pipe = {pipe.sock} from {repr(pipe.route)} using {strategy}.")
                return [pipe, strategy]

        log(f"> Uh oh, pipe was None.")
        return [pipe, None]

    # Main function for scheduling p2p connections.
    async def pipe(self, dest_bytes, timeout=60, strategies=P2P_STRATEGIES):
        # Identify a pipe.
        strategy = p2p_pipe = None
        pipe_id = rand_plain(15)
        assert(isinstance(pipe_id, bytes))

        # Wrap open pipe in a task so it can be cancelled.
        task = asyncio.create_task(
            async_wrap_errors(
                self.open_pipe(pipe_id, dest_bytes, strategies)
            )
        )

        # Record the task in a pending state.
        self.node.pending_pipes[pipe_id] = task
        try:
            await asyncio.wait_for(
                task,
                timeout
            )
            p2p_pipe, strategy = task.result()
        except Exception:
            log_exception()

        # When it's done delete the pending reference.
        if pipe_id in self.node.pending_pipes:
            del self.node.pending_pipes[pipe_id]
            
        return [p2p_pipe, strategy]

    """
    (1) Connects to every available remote address concurrently.
    (2) Chooses the first connection that succeeds.
    (3) Closes any unneeded connections.
    (4) Only connections to an address if the AF is supported.

                 (currently 3) * (valid af no = ip4, ip6)
    max cons = PEER_ADDR_MAX_INTERFACES * 2 
    """
    async def direct_connect(self, p2p_dest, pipe_id, signal_pipe=None, proto=TCP):
        tasks = []
        out = b"ID %s" % (pipe_id)
        for af in VALID_AFS:
            # Check an interface exists for that AF.
            if not len(self.node.ifs.by_af[af]):
                continue

            # Peer doesn't support af type.
            if not len(p2p_dest[af]):
                continue

            async def attempt_con(af, info):
                # (1) Get first interface for AF.
                # (2) Build a 'route' from it with it's main NIC IP.
                # (3) Bind to the route at port 0. Return itself.
                route = await self.node.ifs.get(af).route(af).bind()

                # Connect to this address.
                dest = await Address(
                    str(info["ext"]),
                    info["port"],
                    route
                ).res()

                return await pipe_open(
                    route=route,
                    proto=proto,
                    dest=dest,
                    msg_cb=self.node.msg_cb
                )

            # Similar to 'happy eyeballs.'
            for info in p2p_dest[af]:
                # Schedule the coroutine.
                tasks.append(
                    async_wrap_errors(
                        attempt_con(af, info)
                    )
                )

        # Do all connections concurrently.
        pipes = []
        if len(tasks):
            pipes = await asyncio.gather(*tasks)
            pipes = strip_none(pipes)

        # Only keep the first connection.
        if len(pipes) > 1:
            log("Closing unneeded pipes in direct connect.")
            for i in range(1, len(pipes)):
                await pipes[i].close()

        # On same host recipient needs time to setup con handlers.
        # This prevents race conditions in receiving the con ID.
        await asyncio.sleep(0.1)

        # Send con ID -- used to tell peer which request this relates to.
        if len(pipes):
            log(f"Got pipe in p2p direct {pipes[0].sock}")
            log(f"pipes[0].stream.dest.tup = {pipes[0].stream.dest_tup}")
            log(f"data to send {out}")
            r = await pipes[0].send(out, pipes[0].stream.dest_tup)
            log(f"sending pipe id ret = {r} where 0 is failure.")
            return pipes[0]
        else:
            log("all pipes in direct con returned none.")

        # No cons = failure.
        return None

    """
    Peer A and peer B both have a list of interface info
    for each address family. Each interface connected to
    a unique gateway will have its own NAT characteristics.
    The challenge is to prioritize which combination of
    peer interfaces is most favourable given the interplay
    between their corrosponding NAT qualities.

    The least restrictive NATs are assigned a lower
    value. So the start of the algorithm just groups together
    the lowest value pairs and runs through them until one
    succeeds.
    """
    async def tcp_hole_punch(self, p2p_dest, pipe_id, signal_pipe):
        log("In pipe pipe tcp hole punmch.")
        # [af][addr_index] = [[nat_type, if_info], ...]
        nat_pairs = {}

        # Our address and their address will be sorted.
        addr_list = [self.node.p2p_addr, p2p_dest]

        # Check all valid addresses.
        for af in VALID_AFS:
            # Determine if both sides support the AF.
            skip_af = False
            for addr in addr_list:
                if not len(addr[af]):
                    skip_af = True
                    break

            # Skip an AF if both sides don't support it.
            if skip_af:
                log(f"Skipping AF = {af}")
                continue

            # Create a list of NAT types indexed by addr index.
            nat_pairs[af] = {}
            for addr_index, addr in enumerate(addr_list):
                # Store subset of interface details.
                nat_pairs[af][addr_index] = []

                # Loop over all the interface details by address family.
                for _, if_info in enumerate(addr[af]):
                    # Save interface details we're interested in.
                    nat_pairs[af][addr_index].append([
                        if_info["nat"]["type"],
                        if_info
                    ])

                # Sort the details list based on the first field (NAT type.)
                nat_pairs[af][addr_index] = sorted(
                    nat_pairs[af][addr_index],
                    key=lambda x: x[0]
                )

            # Step through interface details for both addresses.
            our_offset = 0
            their_offset = 0
            while 1:
                # List of some interface details for our addr.
                our_info = nat_pairs[af][0][our_offset][1]

                # List of some interface details for their addr.
                their_info = nat_pairs[af][1][their_offset][1]

                # Load info for hole punching.
                if_index = our_info["if_index"]
                interface = self.node.if_list[if_index]
                stun_client = STUNClient(interface=interface, af=af)
                initiator = self.node.tcp_punch_clients[if_index]

                # Calculate punch mode
                route = interface.route(af)
                their_addr = await Address(
                    str(their_info["ext"]),
                    80,
                    route
                ).res()
                punch_mode = initiator.get_punch_mode(their_addr)

                log(f"Loaded punc mode = {punch_mode}")
                if punch_mode == TCP_PUNCH_REMOTE:
                    use_addr = str(their_info["ext"])
                else:
                    use_addr = str(their_info["nic"])
                log(f"using addr = {use_addr}")
                
                # Step 1 -- set initial mappings for initiator.
                punch_ret = await initiator.proto_send_initial_mappings(
                    use_addr,
                    their_info["nat"],
                    p2p_dest["node_id"],
                    pipe_id,
                    stun_client,
                    mode=punch_mode
                )

                # Build first (required) punch message for peer.
                out = build_punch_response(
                    b"INITIAL_MAPPINGS",
                    pipe_id,
                    punch_ret,
                    self.node.addr_bytes,
                    af,

                    # Which iface we're using from our addr.
                    our_info["if_index"],

                    # Which iface they should use.
                    their_info["if_index"]
                )
                log(f"Sending {out}")

                # Send first protocol signal message to peer.
                send_task = asyncio.create_task(
                    async_wrap_errors(
                        signal_pipe.send_msg(
                            out,
                            to_s(p2p_dest["node_id"])
                        )
                    )
                )

                # Allow for time to receive updated mappings.
                # Peer checks every 2 seconds.
                try:
                    await asyncio.wait_for(
                        punch_ret[2].wait(),
                        4
                    )
                    log("punch: updated mappings received.")
                except asyncio.TimeoutError:
                    # Updated mappings are not always required.
                    log("punch: timed out getting updated mappings.")
                    pass

                # Start task to get a TCP hole.
                pipe = await get_tcp_hole(
                    PUNCH_INITIATOR,
                    pipe_id,
                    p2p_dest["node_id"],
                    initiator,
                    self.node
                )

                # Exit loops if success.
                if pipe is not None:
                    return pipe

                """
                The next pairing is done by increasing their index by 1.
                If it doesn't point to a new element for them then we're done.
                On the other hand if there is a new element the code
                decides whether to increase our offset indicating there
                are more interfaces to try on our side too.

                It may make sense for every one of their interfaces to
                try punch through to every one of our best interfaces.
                But for now they are only tried once.
                    E.g. m vs m * n
                """
                their_offset += 1

                # Don't increase this if there's no new entry.
                if our_offset < len(nat_pairs[af][0]) - 1:
                    our_offset += 1

                # Exit if their offset exceeds the end.
                if their_offset >= len(nat_pairs[af][1]) - 1:
                    break

    async def udp_relay(self, p2p_dest, pipe_id, signal_pipe):
        # Setup a new TURN client.
        # Save it to the list of turn clients.
        async def setup_turn_client(turn_server, interface, af):
            # Resolve TURN domain.
            route = await interface.route(af).bind()
            turn_addr = await Address(
                turn_server["host"],
                turn_server["port"],
                route
            ).res()

            # Build TURN client.
            turn_client = TURNClient(
                route=route,
                turn_addr=turn_addr,
                turn_user=turn_server["user"],
                turn_pw=turn_server["pass"],
                turn_realm=turn_server["realm"],
                msg_cb=self.node.msg_cb
            )

            # Start TURN client.
            await turn_client.start()
            self.node.turn_clients.append(turn_client)
            return turn_client

        # Notify peer to use a TURN server.
        # Returns the turn client on success.
        async def send_turn_request(af, turn_server, turn_client):
            # Save the external address to ensure replies are valid.
            self.node.expected_addrs[pipe_id] = [p2p_dest[af][0]["ext"]]

            # Get our tups to send.
            client_tup = await turn_client.client_tup_future
            relay_tup = await turn_client.relay_tup_future

            # Make a source address to send.
            af = turn_client.turn_addr.af
            interface = turn_client.turn_pipe.route.interface
            if_index = self.node.if_list.index(interface)
            turn_server_index = TURN_SERVERS.index(turn_server)
            turn_client_index = self.node.turn_clients.index(turn_client)
            src_addr_bytes = make_peer_addr(
                self.node.node_id,
                [interface],
                list(self.node.signal_pipes),
                ip=to_b(client_tup[0]),
                port=client_tup[1],
                if_index=if_index
            )

            # Send relay message to peer.
            turn_server = turn_client.get_turn_server()
            out = b"TURN_REQUEST %s %d %d %d %s %s %d %s %d %d %d" % (
                pipe_id,
                af,
                if_index,
                p2p_dest[af][0]["if_index"],
                src_addr_bytes,
                to_b(client_tup[0]),
                int(client_tup[1]),
                to_b(relay_tup[0]),
                int(relay_tup[1]),
                int(turn_server_index),
                int(turn_client_index)
            )
            
            # Send out direct con msg using the above addr.
            node_id = to_s(p2p_dest["node_id"])
            turn_client.new_node_event(node_id)
            await signal_pipe.send_msg(
                to_s(out),
                node_id
            )

            # Return TURN client if it's setup successfully.
            try:
                await asyncio.wait_for(
                    turn_client.node_events[node_id].wait(),
                    10
                )

                self.node.pipes[pipe_id] = turn_client
                return turn_client
            except asyncio.TimeoutError:
                log("> pipe turn req timed out.")
                return None

        # See if a compatible TURN client already exists.
        async def reuse_turn_client(turn_servers, retry_no):
            i = 0
            for turn_client in self.node.turn_clients:
                # Find related setting for turn_client.
                turn_info = None
                for turn_server in turn_servers:
                    found_client = self.node.find_turn_client(turn_server)
                    if found_client == turn_client:
                        turn_info = turn_server
                        break

                # Could not find the server's settings.
                if turn_info is None:
                    break

                # Find a supported AF.
                for af in turn_info["afs"]:
                    # Other peer can't handle this address family.
                    if not len(p2p_dest[af]):
                        continue

                    # Use this TURN client.
                    out = await send_turn_request(af, turn_info, turn_client)
                    if out is None:
                        i += 1
                        continue

                    return out

                # Existing clients failed.
                if i >= retry_no:
                    break

        # Create a new TURN client for use with a peer.
        # Ensure the chosen TURN server supports the AF of ourself + peer.
        async def use_new_turn_client(turn_servers, retry_no):
            i = 0
            for turn_server in turn_servers:
                for interface in self.node.if_list:
                    for af in turn_server["afs"]:
                        # Too many attempts.
                        if i >= retry_no:
                            return None
                        
                        # Other peer can't handle this address family.
                        if not len(p2p_dest[af]):
                            continue

                        # Interface supports AF.
                        if af not in interface.supported():
                            continue

                        # Make a new TURN client.
                        turn_client = await async_wrap_errors(
                            setup_turn_client(
                                turn_server=turn_server, 
                                interface=interface,
                                af=af
                            ),
                            timeout=5
                        )
                        
                        # Exception occured.
                        if turn_client is None:
                            i += 1
                            continue

                        # Notify the peer of the client.
                        out = await send_turn_request(af, turn_server, turn_client)
                        if out is None:
                            i += 1
                            continue

                        return out

        # Shuffle the list of TURN servers for load balancing.
        # Prioritize using p2pd as the first attempt.
        retry_no = 3
        turn_servers = [TURN_SERVERS[0]] + shuffle(TURN_SERVERS[1:])

        # Try reuse an existing TURN client if one is already compatible.
        out = await reuse_turn_client(turn_servers, retry_no)
        if out is None:
            # Create a new TURN client and use that.
            return await use_new_turn_client(turn_servers, retry_no)
        else:
            # Got a valid client -- return reference.
            return out

if __name__ == "__main__": # pragma: no cover
    async def test_p2p_con():
        p2p_dest = None
        if1 = await Interface("enp3s0").start()
        if_list = [if1]
        pipe_id = rand_plain(10)
        p2p_pipe = P2PPipe(if_list, 0)
        await p2p_pipe.direct_connect(p2p_dest, pipe_id, proto=TCP)


        while 1:
            await asyncio.sleep(1)

    async_test(test_p2p_con)


"""
The UPnP module implements IPv4 port forwarding (for NATs) and for IPv6
it provides the means to add exceptions to the router's firewall (which
it calls adding 'pin holes'.)

When using IPv6: the address to make an exception for must be the one
used to make the requests (this is handled automatically here.)
Therefore the interface used must be able to bind to the address
being added as an exception. IPv6 also requires these exceptions have
an expiry which maxes out at 24 hours (it's expressed in seconds.)
Thus, a worker process to re-add the exception at expiry is a good idea.

In IPv4 a lease time of 0 (unlimited) is allowed which is the approach
taken here. So duel-stack hosts will at least have one route that's
reachable. Assuming that the rules aren't wiped out after the router
is rebooted. It's quite possible they are. Maybe useful for security
and cleanup purposes. 

Finally, the response messages after port mapping can be inconsistent
with the true outcome of the request. In testing IPv4 port forwarding
on an Open-WRT VirtualBox VM using the miniupnpd package it replies
with a 501 error message when follow-up calls indicate the mappings
were created successfully. UPnP stacks aren't perfect.

Developer resources: 
https://github.com/jeremypoulter/DeveloperToolsForUPnP
    - The AV server does IPv6 and is useful for testing IPv6 code.
https://openwrt.org/docs/guide-user/virtualization/vmware#upgradedupdated_ova_for_openwrt21
    - This is a massive guide on how to get Open-WRT to run in VMWare.
    Do not follow the first part. Skip directly to the section that
    has an 'updated OVA for VMWare' this file is gold.
    Don't use VMWare for it. Open this in VirtualBox.
    It is configured to use LAN IP 192.168.1.1 by default. 

    This is important:
        - When you start the VM enter passwd and set a password for root
        - enter vi /etc/config/network
        - change 192.168.1.1 to an IP in the same subnet as the
        network interface you're using for the Internet.
        (press esc then i for insert mode. esc then :qw enter to save/quit.)
        - enter reboot to restart the VM.
        - You should now be able to visit that IP in a web browser.
        - Enter the root password you set or try blank.
        - Go to the software section and update it.
        - Install miniupnpd -- the config is at /etc/config/upnpd
        - Enable it in the config file and reboot again.
http://upnp.org/specs/gw/UPnP-gw-WANIPv6FirewallControl-v1-Service.pdf
    - Specification for add pin hole for ipv6
https://github.com/PortSwigger/upnp-hunter/blob/master/UPnPHunter_Burp.py
    - Useful reference for SSPD search code
https://www.rapid7.com/blog/post/2020/12/22/upnp-with-a-holiday-cheer/
https://stackoverflow.com/questions/54802371/upnp-ssdp-discovery-with-ipv6
    - Programming references on UPnP port forwarding mostly
https://community.ui.com/questions/Ports-required-for-upnp2/6692d89e-1dd6-4abd-a6fa-350cf3444832
    - Reference for some default ports for SSDP services.
"""

import socket
import xmltodict
from .utils import *
from .net import *
from .address import *
from .base_stream import *
from .http_client_lib import *

UPNP_LEASE_TIME = 86399
UPNP_PORT = 1900
UPNP_IP   = {
    IP4: b"239.255.255.250",
    IP6: b"FF02::C"
}

"""
Port forwarding should only be done once per interface
address but each interface needs its own unique port.
Otherwise there would be conflicts with past mappings.
The function here associates a unique port with an
interface IP from a high order port range to bind on.
It's not perfects but it keeps post allocations
deterministic yet somewhat unique per NIC IP.
"""
def get_port_by_ips(if_names, port_range=[10000, 65000]):
    # Convert IP to int.
    if_names = sorted(if_names)
    as_str = ""
    for if_name in if_names:
        as_str += if_name
    as_int = hash(as_str)

    # Wrap int around finite field.
    port = field_wrap(as_int, port_range)

    return port

def m_search_buf(af):
    if af == IP4:
        host = to_s(UPNP_IP[af])
    if af == IP6:
        host = f"[{to_s(UPNP_IP[af])}]"

    buf = \
    f'M-SEARCH * HTTP/1.1\r\n' \
    f'HOST: {host}:{UPNP_PORT}\r\n' \
    f'ST: upnp:rootdevice\r\n' \
    f'MX: 5\r\n' \
    f'MAN: "ssdp:discover"\r\n' \
    f'\r\n'

    return to_b(buf)

def parse_root_xml(d, service_type):
    results = []
    for k, v in d.items():
        if isinstance(v, list):
            for e in v:
                results += parse_root_xml(e, service_type)
        elif isinstance(v, dict):
            results += parse_root_xml(v, service_type)
        else:
            if k == "serviceType":
                if service_type in v:
                    results.append(d)
                    break
    #
    return results

def port_forward_task(route, dest, service, lan_ip, lan_port, ext_port, proto, desc):
    # Do port forwarding.
    desc = to_s(desc)
    if dest.af == IP4:
        soap_action = "AddPortMapping"
        body = f"""
<u:{soap_action} xmlns:u="{service["serviceType"]}">
    <NewRemoteHost></NewRemoteHost>
    <NewExternalPort>{ext_port}</NewExternalPort>
    <NewProtocol>{proto}</NewProtocol>
    <NewInternalPort>{lan_port}</NewInternalPort>
    <NewInternalClient>{lan_ip}</NewInternalClient>
    <NewEnabled>1</NewEnabled>
    <NewPortMappingDescription>{desc}</NewPortMappingDescription>
    <NewLeaseDuration>0</NewLeaseDuration>
</u:{soap_action}>
        """

    # Add a hole in the firewall.
    # Have not added UniqueID -- will it still work?
    if dest.af == IP6:
        # Protocol field based on IANA protocol numbers.
        proto_no = proto
        if proto.lower() == "tcp":
            proto_no = 6
        if proto.lower() == "udp":
            proto_no = 17

        # https://github.com/miniupnp/miniupnp/issues/228
        soap_action = "AddPinhole"
        body = f"""
<u:{soap_action} xmlns:u="{service["serviceType"]}">
    <RemoteHost></RemoteHost>
    <RemotePort>0</RemotePort>
    <Protocol>{proto_no}</Protocol>
    <InternalPort>{lan_port}</InternalPort>
    <InternalClient>{lan_ip}</InternalClient>
    <LeaseTime>{UPNP_LEASE_TIME}</LeaseTime>
</u:{soap_action}>
        """

    # Build the XML payload to send.
    payload = f"""
<?xml version="1.0"?>
<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
<s:Body>
{body}
</s:Body>
</s:Envelope>
    """

    # Custom headers for soap.
    headers = [
        [
            b"SOAPAction",
            to_b(f"\"{service['serviceType']}#{soap_action}\"")
        ],
        [b"Connection", b"Close"],
        [b"Content-Type", b"text/xml"],
        [b"Content-Length", to_b(f"{len(payload)}")]
    ]

    return http_req(
        route,
        dest,
        service["controlURL"],
        do_close=True,
        method="POST",
        payload=payload,
        headers=headers
    )

async def add_fixed_paths(interface, af, get_root_desc):
    # List of hosts to try get a rootXML from.
    hosts = []
    gws = interface.netifaces.gateways()[af]

    # Add all gateways netiface knows about.
    if len(gws):
        for gw in gws:
            hosts.append(gw[0])

    # Valid default gateway address in IPv6.
    if af == IP6:
        hosts.append("FE80::1")

    # Nothing to do.
    if not len(hosts):
        return []

    # Ports to try.
    ports = [
        # MiniUPnP
        5000,

        # Libupnp
        49152,

        # Many routers
        5431,

        # Default web server ports.
        80,
        8080
    ]

    # List of control urls to try.
    paths = [
        "/rootDesc.xml",
        "/"
    ]

    # Used to carry out the work.
    tasks = []
    async def worker(host, port, path):
        route = await interface.route(af).bind()
        dest = await Address(host, port, route).res()
        return await get_root_desc(route, dest, path)

    # Build list of tasks.
    for host in hosts:
        for port in ports:
            for path in paths:
                tasks.append(
                    async_wrap_errors(
                        worker(host, port, path)
                    )
                )

    # Execute fetch attempts.
    results = await asyncio.gather(*tasks)
    out = []
    for result in results:
        if len(result):
            out += result

    return out

async def port_forward(interface, ext_port, src_addr, desc, proto="TCP"):
    # Set protocol family for multicast socket.
    af = src_addr.af
    if af == IP4:
        sock_conf = dict_child({
            "sock_proto": socket.IPPROTO_UDP
        }, NET_CONF)
    if af == IP6:
        sock_conf = dict_child({
            "sock_proto": socket.IPPROTO_UDP
        }, NET_CONF)

    # Make multicast socket for M-search.
    route = interface.route(af)
    await route.bind(ips=route.nic())
    sock = await socket_factory(route, sock_type=UDP, conf=sock_conf)
    if af == IP6:
        sock.setsockopt(socket.IPPROTO_IPV6, socket.IP_MULTICAST_TTL, 2)

    # Create async pipe wrapper for multicast socket.
    dest = await Address(UPNP_IP[af], UPNP_PORT, route).res()
    pipe = await pipe_open(UDP, route, dest, sock, conf=sock_conf)
    pipe.subscribe()

    # Send m-search message.
    buf = m_search_buf(af)

    # Multiple sends spaced apart because UDP is garbage.
    for i in range(0, 3):
        await pipe.send(buf)
        await asyncio.sleep(0.1)

    # Get list of HTTP replies from M-Search message.
    replies = []
    while 1:
        out = await pipe.recv(timeout=4)
        if out is None:
            break

        try:
            reply = ParseHTTPResponse(out)
        except Exception:
            continue

        replies.append(reply)

    # Cleanup multicast socket.
    await pipe.close()

    # Filter duplicate replies.
    unique = {}
    for reply in replies:
        if "location" not in reply.hdrs:
            continue

        location = reply.hdrs["location"]
        if location in unique:
            continue

        unique[location] = reply

    # Save only unique replies.    
    replies = list(unique.values())

    # Main code that gets a list of port forward tasks for a device.
    async def get_root_desc(route, dest, path):
        # Service type lookup table.
        service_types = {
            IP4: "WANIPConnection",
            IP6: "WANIPv6FirewallControl"
        }

        # Bind ip for http forwarding.
        if src_addr.af == IP6:
            # The source address needs to match the internal client
            # being forwarded for add pin hole.
            ips = src_addr.tup[0]
        else:
            # The source address doesn't matter in IPv6 port forwarding.
            ips = src_addr.tup[0]

        # Get main XML for device.
        tasks = []
        try:
            # Request rootDesc.xml.
            _, http_resp = await http_req(route, dest, path, do_close=True)
            if http_resp is None:
                return []

            xml = http_resp.out()
            d = xmltodict.parse(xml)

            # Convert to a list of services.
            services = parse_root_xml(d, service_types[route.af])
        except Exception:
            log(f"Failed to get root xml {dest.tup} {path}")
            log_exception()
            return tasks

        # Get a task to port forward for each service.
        for service in services:
            forward_route = await interface.route(route.af).bind(ips=ips)
            task = async_wrap_errors(
                port_forward_task(
                    forward_route,
                    dest,
                    service,
                    src_addr.tup[0],
                    src_addr.tup[1],
                    ext_port,
                    proto,
                    desc
                )
            )
            
            # Save task.
            tasks.append(task)

        return tasks

    """
    for reply in replies:
        if "location" not in reply.hdrs:
            continue

        print(reply.hdrs["location"])
    """

    # Port forward on all devices that replied.
    tasks = []
    for req in replies:
        # Location header points to rootDesc.xml.
        url = urllib.parse.urlparse(req.hdrs["location"])
        hostname = url.hostname
        if src_addr.af == IP6:
            hostname = hostname.strip("[]")

        # Root XML address.
        xml_dest = await Address(
            hostname,
            url.port,
            route
        ).res()

        # Request rootDesc.xml.
        http_route = await interface.route(af).bind(ips=src_addr.tup[0])
        results = await async_wrap_errors(
            get_root_desc(http_route, xml_dest, url.path)
        )

        if results is None:
            continue
        
        if len(results):
            tasks += results

    # Generate fixed path attempts to increase success if m-search fails.
    tasks += await add_fixed_paths(interface, af, get_root_desc)

    # Return port forward tasks.
    return tasks

if __name__ == "__main__":
    async def upnp_main():
        from .interface import Interfaces, init_p2pd
        i = await Interface().start_local()
        port = 31375
        desc = b"test 10003"


        print(i.route(IP4).nic())

        # respondes dont indicate success.

        class F():
            def __init__(self, i, af):
                self.interface = i
                self.af = af

        f = F(i, IP4)
        #src_addr = await Address("fe80::1131:b51a:3f8f:1f2d", port).res(f)
        src_addr = await Address("192.168.21.21", port, f).res()
        tasks = await port_forward(i, port, src_addr, desc)
        results = await asyncio.gather(*tasks)

        print(results)

        for r in results:
            print(r[1].out())

        return

        x = results[0][1].out()
        print(x)

    async_test(upnp_main)




"""
This module is a simple port of the Sqlite-based key-value store
described in this Stackoverflow answer:
https://stackoverflow.com/questions/47237807/use-sqlite-as-a-keyvalue-store

The port simply replaces blocking functions with async calls using aiosqlite.
This offers a few benefits over existing open source modules:

    - It works on very old versions of Python (>= 3.6)
    - It plays well with concurrency and asyncio. 
    - It's simpler than having to define a schema.
    - And finally: it automatically converts Python data types.

Using Sqlite as the underlying library also makes the database highly reliable,
resistent to corruption, and probably has safe-guards for multi-process access.
"""

import asyncio
from collections import OrderedDict
import aiosqlite

class SqliteKVS():
    def __init__(self, file_path):
        self.file_path = file_path

    # Scheme has a text key and value.
    async def start(self):
        query = "CREATE TABLE IF NOT EXISTS kv" + \
                "(key text unique, value text)"
        
        self.db = await aiosqlite.connect(self.file_path)
        await self.db.execute(query)
        await self.db.commit()
        return self

    # Ensure the DB is closed to avoid corruption.
    async def close(self):
        await self.db.close()
        self.db = None

    async def __aenter__(self):
        return await self.start()
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def len(self):
        query = 'SELECT COUNT(*) FROM kv'
        cursor = await self.db.execute(query)
        rows = (await cursor.fetchone())[0]
        await cursor.close()
        return rows if rows is not None else 0

    # Overwrite a row attached to a key.
    # Create it if it doesn't exist.
    # Store complex objects as text.
    async def put(self, key, value):
        query = 'REPLACE INTO kv (key, value) VALUES (?,?)'
        value = repr(value)
        await self.db.execute(query, (key, value))
        await self.db.commit()

    # Select values by key name.
    # Decode any complex values back to Python types.
    async def get(self, key, default=None):
        query = 'SELECT value FROM kv WHERE key = ?'
        cursor = await self.db.execute(query, (key,))
        item = await cursor.fetchone()
        await cursor.close()

        # No entry by that key exists.
        if item is None:
            if default is not None:
                return default
            raise KeyError(key)
        
        return eval(item[0])
    
    # Delete key entry in DB.
    async def unset(self, key):
        # Raise exception if it doesn't exist.
        await self.get(key)

        # Delete the key.
        query = 'DELETE FROM kv WHERE key = ?'
        await self.db.execute(query, (key,))
        await self.db.commit()

    async def iterkeys(self):
        query = 'SELECT key FROM kv'
        async with self.db.execute(query) as cursor:
            async for row in cursor:
                yield row[0]

    async def itervalues(self):
        query = 'SELECT value FROM kv'
        async with self.db.execute(query) as cursor:
            async for row in cursor:
                yield eval(row[0])

    async def iteritems(self):
        query = 'SELECT key, value FROM kv'
        async with self.db.execute(query) as cursor:
            async for row in cursor:
                yield row[0], eval(row[1])

    async def keys(self):
        key_list = []
        async for key in self.iterkeys():
            key_list.append(key)

        return key_list
    
    async def values(self):
        value_list = []
        async for value in self.itervalues():
            value_list.append(value)

        return value_list
    
    async def items(self):
        item_list = []
        async for item in self.iteritems():
            item_list.append(item)

        return item_list
    
    async def contains(self, key):
        query = 'SELECT 1 FROM kv WHERE key = ?'
        cursor = await self.db.execute(query, (key,))
        ret = await cursor.fetchone()
        await cursor.close()
        return ret is not None
    
    async def __aiter__(self):
        return self.iterkeys()

async def test_sqlite_kvs():
    kvs = SqliteKVS("test_sqlite_db.sqlite")
    await kvs.start()

    await kvs.put("my_test_key", [("test"), {'a': b'b'}])
    await kvs.put("another key", 2)

    #ret = await kvs.get("my_test_key")
    #print(ret)

    async for row in kvs.iterkeys():
        print(row)

    l = await kvs.len()
    print(l)

    r = await kvs.keys()
    print(r)

    r = await kvs.values()
    print(r)

    i = await kvs.items()
    print(i)

    await kvs.put("key_to_dele", 1000)
    print(await kvs.items())
    await kvs.unset("key_to_dele")
    print(await kvs.items())

    print(await kvs.contains("some bs"))
    print(await kvs.contains("my_test_key"))

    async for row in kvs:
        print(row)

    await kvs.close()

    async with SqliteKVS("test_sqlite_db.sqlite") as kvs:
        ret = await kvs.get("my_test_key")
        print(ret)

#async_test(test_sqlite_kvs)


import socket
IP4 = socket.AF_INET
IP6 = socket.AF_INET6

ENABLE_STUN = True
ENABLE_UDP = True
P2PD_TEST_INFRASTRUCTURE = False

"""
To keep things simple P2PD uses a number of services to
help facilitate peer-to-peer connections. At the moment
there is no massive list of servers to use because
(as I've learned) -- you need to also have a way to
monitor the integrity of servers to provide high-quality
server lists to peers. That would be too complex to provide
starting out so this may be added down the road.

Note to any engineers:

If you wanted to run P2PD privately you could simply
point all of these servers to your private infrastructure.
"""


NET_DEBUG_PROVIDERS = {
    IP4: [
        "http://net-debug.000webhostapp.com/net_debug.php"
    ],
    IP6: [
        "http://net-debug.000webhostapp.com/net_debug.php"
    ]
}

NAME_STORE_PROVIDERS = [
    {
        "afs": [IP4, IP6],
        "url": "http://net-debug.000webhostapp.com/name_store.php"
    }
]

"""
Used to lookup what a nodes IP is and do NAT enumeration.
Supports IPv6 / IPv4 / TCP / UDP -- change IP and port requests.

STUNT servers support TCP.
STUND servers support UDP.
"""
STUNT_SERVERS = {
    IP4: [
        {
            "host": "stunserver.stunprotocol.org",
            "primary": {"ip": "3.132.228.249", "port": 3478},
            "secondary": {"ip": "3.135.212.85", "port": 3479},
        },
        {
            "host": "stun.hot-chilli.net",
            "primary": {"ip": "49.12.125.53", "port": 3478},
            "secondary": {"ip": None, "port": None},
        },
        {
            "host": "stun.voip.blackberry.com",
            "primary": {"ip": "20.15.169.8", "port": 3478},
            "secondary": {"ip": None, "port": None},
        },
        {
            "host": "webrtc.free-solutions.org",
            "primary": {"ip": "94.103.99.223", "port": 3478},
            "secondary": {"ip": None, "port": None},
        },
        {
            "host": "stun.siptrunk.com",
            "primary": {"ip": "23.21.92.55", "port": 3478},
            "secondary": {"ip": None, "port": None},
        },
    ],
    IP6: [
        {
            "host": "stunserver.stunprotocol.org",
            "primary": {"ip": "2600:1f16:8c5:101:80b:b58b:828:8df4", "port": 3478},
            "secondary": {"ip": "2600:1f16:08c5:0101:6388:1fb6:8b7e:00c2", "port": 3479},
        }
    ]
}



STUND_SERVERS = {
    IP4: [
        {
            "host": "stun.voipcheap.co.uk",
            "primary": {"ip": "77.72.169.211", "port": 3478},
            "secondary": {"ip": "77.72.169.210", "port": 3479},
        },
        {
            "host": "stunserver.stunprotocol.org",
            "primary": {"ip": "3.132.228.249", "port": 3478},
            "secondary": {"ip": "3.135.212.85", "port": 3479},
        },
        {
            "host": "stun.usfamily.net",
            "primary": {"ip": "64.131.63.216", "port": 3478},
            "secondary": {"ip": "64.131.63.240", "port": 3479},
        },
        {
            "host": "stun.ozekiphone.com",
            "primary": {"ip": "216.93.246.18", "port": 3478},
            "secondary": {"ip": "216.93.246.15", "port": 3479},
        },
        {
            "host": "stun.voipwise.com",
            "primary": {"ip": "77.72.169.213", "port": 3478},
            "secondary": {"ip": "77.72.169.212", "port": 3479},
        },
        {
            "host": "stun.mit.de",
            "primary": {"ip": "62.96.96.137", "port": 3478},
            "secondary": {"ip": "62.96.96.138", "port": 3479},
        },
        {
            "host": "stun.hot-chilli.net",
            "primary": {"ip": "49.12.125.53", "port": 3478},
            "secondary": {"ip": "49.12.125.24", "port": 3479},
        },
        {
            "host": "stun.cheapvoip.com",
            "primary": {"ip": "77.72.169.212", "port": 3478},
            "secondary": {"ip": "77.72.169.213", "port": 3479},
        },
        {
            "host": "webrtc.free-solutions.org",
            "primary": {"ip": "94.103.99.223", "port": 3478},
            "secondary": {"ip": "94.103.99.224", "port": 3479},
        },
        {
            "host": "stun.t-online.de",
            "primary": {"ip": "217.0.12.17", "port": 3478},
            "secondary": {"ip": "217.0.12.18", "port": 3479},
        },
        {
            "host": "stun.sipgate.net",
            "primary": {"ip": "217.10.68.152", "port": 3478},
            "secondary": {"ip": "217.116.122.136", "port": 3479},
        },
        {
            "host": "stun.voip.aebc.com",
            "primary": {"ip": "66.51.128.11", "port": 3478},
            "secondary": {"ip": "66.51.128.12", "port": 3479},
        },
        {
            "host": "stun.callwithus.com",
            "primary": {"ip": "158.69.57.20", "port": 3478},
            "secondary": {"ip": "149.56.23.84", "port": 3479},
        },
        {
            "host": "stun.counterpath.net",
            "primary": {"ip": "216.93.246.18", "port": 3478},
            "secondary": {"ip": "216.93.246.15", "port": 3479},
        },
        {
            "host": "stun.ekiga.net",
            "primary": {"ip": "216.93.246.18", "port": 3478},
            "secondary": {"ip": "216.93.246.15", "port": 3479},
        },
        {
            "host": "stun.internetcalls.com",
            "primary": {"ip": "77.72.169.212", "port": 3478},
            "secondary": {"ip": "77.72.169.213", "port": 3479},
        },
        {
            "host": "stun.voipbuster.com",
            "primary": {"ip": "77.72.169.212", "port": 3478},
            "secondary": {"ip": "77.72.169.213", "port": 3479},
        },
        {
            "host": "stun.12voip.com",
            "primary": {"ip": "77.72.169.212", "port": 3478},
            "secondary": {"ip": "77.72.169.213", "port": 3479},
        },
        {
            "host": "stun.freecall.com",
            "primary": {"ip": "77.72.169.211", "port": 3478},
            "secondary": {"ip": "77.72.169.210", "port": 3479},
        },
        {
            "host": "stun.nexxtmobile.de",
            "primary": {"ip": "5.9.87.18", "port": 3478},
            "secondary": {"ip": "136.243.205.11", "port": 3479},
        },
        {
            "host": "stun.siptrunk.com",
            "primary": {"ip": "23.21.92.55", "port": 3478},
            "secondary": {"ip": "34.205.214.84", "port": 3479},
        },
    ],
    IP6: [
        {
            "host": "stunserver.stunprotocol.org",
            "primary": {"ip": "2600:1f16:8c5:101:80b:b58b:828:8df4", "port": 3478},
            "secondary": {"ip": "2600:1f16:08c5:0101:6388:1fb6:8b7e:00c2", "port": 3479},
        },
        {
            "host": "stun.hot-chilli.net",
            "primary": {"ip": "2a01:4f8:242:56ca::2", "port": 3478},
            "secondary": {"ip": "2a01:04f8:0242:56ca:0000:0000:0000:0003", "port": 3479},
        },
        {
            "host": "stun.simlar.org",
            "primary": {"ip": "2a02:f98:0:50:2ff:23ff:fe42:1b23", "port": 3478},
            "secondary": {"ip": "2a02:0f98:0000:0050:02ff:23ff:fe42:1b24", "port": 3479},
        },
    ]
}

"""
stun_test = [["stunserver.stunprotocol.org", 3478]]
STUNT_SERVERS = { IP4: stun_test, IP6: stun_test }
STUND_SERVERS = STUNT_SERVERS
"""

# The main server used to exchange 'signaling' messages.
# These are messages that help nodes create connections.
MQTT_SERVERS = [
    # [b"p2pd.net", 1883],
    [b"mqtt.eclipseprojects.io", 1883],
    [b"broker.mqttdashboard.com", 1883],
    [b"test.mosquitto.org", 1883],
    [b"broker.emqx.io", 1883],
    [b"broker.hivemq.com", 1883],
]

# Port is ignored for now.
NTP_SERVERS = [
    ["time.google.com", 123],
    ["pool.ntp.org", 123],
    ["time.cloudflare.com", 123],
    ["time.facebook.com", 123],
    ["time.windows.com", 123],
    ["time.apple.com", 123],
    ["time.nist.gov", 123],
    ["utcnist.colorado.edu", 123],
    ["ntp2.net.berkeley.edu", 123],
    ["time.mit.edu", 123],
    ["time.stanford.edu", 123],
    ["ntp.nict.jp", 123],
    ["ntp1.hetzner.de", 123],
    ["ntp.ripe.net", 123],
    ["clock.isc.org", 123],
    ["ntp.ntsc.ac.cn", 123],
    ["1.amazon.pool.ntp.org", 123],
    ["0.android.pool.ntp.org", 123],
    ["0.pfsense.pool.ntp.org", 123],
    ["0.debian.pool.ntp.org", 123],
    ["0.gentoo.pool.ntp.org", 123],
    ["0.arch.pool.ntp.org", 123],
    ["0.fedora.pool.ntp.org", 123],
]

"""
These are TURN servers used as fallbacks (if configured by a P2P pipe.)
They are not used for 'p2p connections' by default due to their use of
UDP and unordered delivery but it can be enabled by adding 'P2P_RELAY'
to the strategies list in open_pipe().

Please do not abuse these servers. If you need proxies use Shodan or Google
to find them. If you're looking for a TURN server for your production
Web-RTC application you should be running your own infrastructure and not
rely on public infrastructure (like these) which will be unreliable anyway.

Testing:

It seems that recent versions of Coturn no longer allow you to relay data
from your own address back to yourself. This makes sense -- after-all
-- TURN is used to relay endpoints and it doesn't make sense to be
relaying information back to yourself. But it has meant to designing a
new way to test these relay addresses that relies on an external server
to send packets to the relay address.

Note:
-----------------------------------------------------------------------
These servers don't seem to return a reply on the relay address.
Most likely this is due to the server using a reply port that is different
to the relay port and TURN server port. This will effect most types of 
NATs, unfortunately. So they've been removed from the server list for now.

{
    "host": b"webrtc.free-solutions.org",
    "port": 3478,
    "afs": [IP4],
    "user": b"tatafutz",
    "pass": b"turnuser",
    "realm": None
},


{
    "host": b"openrelay.metered.ca",
    "port": 80,
    "afs": [IP4],
    "user": b"openrelayproject",
    "pass": b"openrelayproject",
    "realm": None
}
    
    {
        "host": b"p2pd.net",
        "port": 3478,
        "afs": [IP4, IP6],
        "user": None,
        "pass": None,
        "realm": b"p2pd.net"
    },
"""
TURN_SERVERS = [
    {
        "host": b"p2pd.net",
        "port": 34780,
        "afs": [IP4, IP6],
        "user": b"",
        "pass": b"",
        "realm": b"p2pd.net"
    },
    {
        "host": b"peerjs.com",
        "port": 3478,
        "afs": [IP4, IP6],
        "user": b"peerjs",
        "pass": b"peerjsp",
        "realm": None
    },
    {
        "host": b"turn.obs.ninja",
        "port": 443,
        "afs": [IP4, IP6],
        "user": b"steve",
        "pass": b"setupYourOwnPlease",
        "realm": None
    },
    {
        "host": b"stun.contus.us",
        "port": 3478,
        "afs": [IP4],
        "user": b"contus",
        "pass": b"SAE@admin",
        "realm": None
    },
    {
        "host": b"turn.quickblox.com",
        "port": 3478,
        "afs": [IP4],
        "user": b"quickblox",
        "pass": b"baccb97ba2d92d71e26eb9886da5f1e0",
        "realm": None
    },
    {
        "host": b"turn.threema.ch",
        "port": 443,
        "afs": [IP4],
        "user": b"threema-angular",
        "pass": b"Uv0LcCq3kyx6EiRwQW5jVigkhzbp70CjN2CJqzmRxG3UGIdJHSJV6tpo7Gj7YnGB",
        "realm": None
    },
]





from p2pd.test_init import *
from p2pd import IPRange, HOST_TYPE_IP, HOST_TYPE_DOMAIN, Address
from p2pd import IP6, IP4, Route, Interface
from p2pd.net import ip_norm, RUDP, UDP, NET_CONF
from p2pd.base_stream import pipe_open, SUB_ALL, BaseProto
from p2pd.ack_udp import ACKUDP, BaseACKProto

class TestRUDP(unittest.IsolatedAsyncioTestCase):
    async def test_rudp(self):
        i = await Interface().start_local()
        af = i.supported()[0]
        port = 40000
        r = await i.route(af).bind(port)
        dest_tup = (r.nic(), port)
        dest = await Address(*dest_tup, r).res()
        pipe = (await pipe_open(
            route=r,
            proto=RUDP,
            dest=dest
        )).subscribe(SUB_ALL)


        msg = b"test meow"
        task, event = await pipe.stream.ack_send(msg, dest_tup)
        await asyncio.wait_for(
            event.wait(),
            2
        )
        out = await pipe.recv(SUB_ALL)
        self.assertEqual(out, msg)
        await pipe.close()

    async def test_is_unique(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        b = BaseProto(sock=s)
        d = b"ayy lmao"
        c = ("127.0.0.1", 1337)
        r = b.is_unique_msg(
            pipe=None,
            data=d,
            client_tup=c
        )

        self.assertTrue(r)
        s.close()



if __name__ == '__main__':
    main()


import asyncio
from gmqtt import Client as MQTTClient
from .net import *
from .settings import *

MQTT_CONF = dict_child({
    "con_timeout": 4,
    "recv_timeout": 4
}, NET_CONF)

async def f_proto_print(data):
    print(data)

class SignalMock():
    def __init__(self, peer_id, f_proto, mqtt_server, conf=MQTT_CONF):
        # Setup.
        self.peer_id = to_s(peer_id)
        self.conf = conf
        self.f_proto = f_proto
        self.sub_ready = asyncio.Event()
        self.mqtt_server = mqtt_server
        self.is_connected = False

        # Other.
        self.client = None

        # Tasks pending.
        self.pending_tasks = []

    def on_message(self, client, topic, payload, qos, properties):
        self.pending_tasks.append(
            asyncio.create_task(
                async_wrap_errors(
                    self.f_proto(payload, self),

                    # Set a timeout of 20 seconds to do tasks.
                    # Make everything timeout and end if it meets this.
                    #20
                )
            )
        )

    def on_connect(self, client, flags, rc, properties):
        self.is_connected = True
        client.subscribe(self.peer_id, qos=2)

    def on_disconnect(self, client, packet, exc=None):
        log("Signal pipe disconnected.")

    def on_subscribe(self, client, mid, qos, properties):
        self.sub_ready.set()

    async def start(self):
        self.client = await self.get_client(self.mqtt_server)
        return self

    async def send(self, data, client_tup):
        await self.send_msg(data, client_tup)

    async def send_msg(self, msg, peer_id):
        log(f"> Send signal to {peer_id} = {msg}.")
        self.client.publish(to_s(peer_id), to_s(msg), qos=2)

    async def echo(self, msg, dest_chan):
        out = f"ECHO {self.peer_id} {msg}"
        await self.send_msg(to_s(out), to_s(dest_chan))

    async def get_client(self, mqtt_server):
        client = MQTTClient(self.peer_id)
        client.set_config({'reconnect_retries': -1, 'reconnect_delay': 60})
        client.on_connect = self.on_connect
        client.on_message = self.on_message
        client.on_disconnect = self.on_disconnect
        client.on_subscribe = self.on_subscribe

        await asyncio.wait_for(
            client.connect(host=mqtt_server[0], port=mqtt_server[1]),
            5
        )

        return client

    async def close(self):
        if self.client is not None:
            await self.client.disconnect()

if __name__ == "__main__": # pragma: no cover
    async def f_proto(msg):
        print(type(msg))
        print(msg)

    async def test_signal_mock():
        peer_id = "sdfjk12j312j312j3qsafd"
        s = await SignalMock(peer_id, f_proto).start()

        await s.send_msg(b"test msg", peer_id)

        while 1:
            await asyncio.sleep(1)

    async_test(test_signal_mock)


import platform
import sys
import os
import multiprocessing
from p2pd.test_init import *
from p2pd.cmd_tools import *


class TestCmd(unittest.IsolatedAsyncioTestCase):
    async def do_esc_test(self, s):
        # So paths with spaces don't break the command.
        py = sys.executable
        c = "import sys; print(sys.argv[-1], end='')"
        buf = f'{py}" -c "{c}" ' + s
        return await cmd(buf)

    async def test_escape(self):
        if platform.system() in ["Windows"]:
            tests = [
                # Regular command with double quotes in it.
                [
                    'test "somèthing"',
                    '"test \\"somèthing\\""'
                ],

                # Sneaky attempt to escape enclosing last slash.
                [
                    'test x\\',
                    '"test x\\\\"'
                ],

                # Another attempt to escape last enclosing slash.
                # As long as it's in double quotes it has no effect.
                [
                    'test x^',
                    '"test x^^"'
                ],

                # Test some special charas.
                [
                    'test ! %',
                    '"test ^! ^%"'

                ],

                # Test escape list impact inside string.
                [
                    ',:;=\t&><|',
                    '"^,:^;=\t^&^>^<^|"'
                ],

                # Test unbalanced double quotes escape.
                [
                    'hax """',
                    '"hax \\"\\"\\""'
                ]
            ]

        if platform.system() == "Linux":
            tests = [
                # Regular command with double quotes in it.
                [
                    'tèst "something"',
                    "'tèst \"something\"'"
                ],

                # Sneaky attempt to escape enclosing last slash.
                [
                    'test x\\',
                    "'test x\\'"
                ],

                # Test some special charas.
                [
                    'test ! %',
                    "'test ! %'"

                ],

                # Test escape list impact inside string.
                [
                    ',:;=\t&><|',
                    "',:;=\t&><|'"
                ],

                # Test unbalanced double quotes escape.
                # IDK why it's escaped like this but seems it works.
                [
                    'hax \'\'\'',
                    '\'hax \'"\'"\'\'"\'"\'\'"\'"\'\''
                ]
            ]

        if platform.system() in ["FreeBSD", "Darwin"]:
            tests = [
                # Regular command with double quotes in it.
                [
                    'tèst "something"',
                    '"tèst \\"something\\""'
                ],

                # Sneaky attempt to escape enclosing last slash.
                [
                    'test x\\',
                    '"test x\\\\"'
                ],

                # Another attempt to escape last enclosing slash.
                # As long as it's in double quotes it has no effect.
                [
                    'test x^',
                    '"test x^"'
                ],

                # Test some special charas.
                [
                    'test ! %',
                    '"test ! %"'

                ],

                # Test escape list impact inside string.
                [
                    ',:;=\t&><|',
                    '",:;=\t&><|"'
                ],

                # Test unbalanced double quotes escape.
                [
                    'hax """',
                    '"hax \\"\\"\\""'
                ]
            ]

        esc = get_arg_escape_func()
        for test in tests:
            unsafe_param, safe_param = test
            out_param = esc(unsafe_param)
            self.assertEqual(out_param, safe_param)

            # How is param serialized.
            exp_param = await self.do_esc_test(safe_param)
            if exp_param != unsafe_param:
                print(f"{exp_param} != {unsafe_param}")

            self.assertTrue(isinstance(exp_param, str))

    async def test_cmd(self):
        py = sys.executable
        out = await cmd(f""""{py}" -c "print('something')" """)
        self.assertTrue("something" in out)

    async def test_is_root(self):
        b = is_root()
        self.assertTrue(b in [True, False])

    async def test_nt_is_admin(self):
        if platform.system() != "Windows":
            return

        b = nt_is_admin()
        self.assertTrue(b in [True, False])

    async def test_nt_is_pshell_restricted(self):
        if platform.system() != "Windows":
            return

        out = await is_pshell_restricted()
        self.assertTrue(out in [True, False])

    async def test_nt_set_pshell_unrestricted(self):
        if platform.system() != "Windows":
            return

        # Process must be elevated to test this.
        if not is_root():
            return

        await nt_set_pshell_unrestricted()
        out = await is_pshell_restricted()
        self.assertTrue(out in (True, False))

if __name__ == '__main__':
    multiprocessing.set_start_method("spawn")
    main()


from p2pd.utils import sorted_search

assert(
    sorted_search(
        [1, 40], 324324
    ) == 1
)

assert(
    sorted_search(
        [1, 40], 40
    ) == 1
)

assert(
    sorted_search(
        [1, 40], 3
    ) == 1
)

assert(
    sorted_search(
        [1, 40], 1
    ) == 0
)

assert(
    sorted_search(
        [1, 40], 0
    ) == 0
)

assert(
    sorted_search(
        [1], 1
    ) == 0
)

assert(
    sorted_search(
        [1], 0
    ) == 0
)


assert(
    sorted_search(
        [5, 10, 15], 20
    ) == 2
)

assert(
    sorted_search(
        [5, 10, 15], 10
    ) == 1
)
assert(
    sorted_search(
        [5, 10, 15], 15
    ) == 2
)

assert(
    sorted_search(
        [5, 10, 15], 5
    ) == 0
)

assert(
    sorted_search(
        [5, 10, 15], 3
    ) == 0
)

assert(
    sorted_search(
        [5, 10, 15], 6
    ) == 1
)

assert(
    sorted_search(
        [5, 10, 15, 20], 20
    ) == 3
)

assert(
    sorted_search(
        [5, 10, 15, 22, 40, 80, 100], 30
    ) == 4
)


import platform
from p2pd.test_init import *
from p2pd import IPRange, HOST_TYPE_IP, HOST_TYPE_DOMAIN
from p2pd import IP6, IP4, Route
from p2pd.net import *

class TestNet(unittest.IsolatedAsyncioTestCase):
    async def test_ip_norm(self):
        tests = [
            ["1.1.1.1%test", "1.1.1.1"],
            ["1.1.1.1/24", "1.1.1.1"],
            ["1.1.1.1%test/24", "1.1.1.1"],
            ["::", ("0000:" * 8)[:-1]],
            ["::%test/24", ("0000:" * 8)[:-1]],
            ["::", ("0000:" * 8)[:-1]],
            ["2402:1f00:8101:083f:0000:0000:0000:0001", "2402:1f00:8101:083f:0000:0000:0000:0001"]
        ]

        for src_ip, out_ip in tests:
            self.assertEqual(ip_norm(src_ip), out_ip)

    async def test_rand_link_local(self):
        out = ipv6_rand_link_local()
        ipaddress.IPv6Address(out)

    async def test_netmask_to_cidr(self):
        nm = "255.255.255.255"
        out = netmask_to_cidr(nm)
        self.assertEqual(32, out)

    async def test_toggle_host_bits(self):
        nm = "255.255.0.0"
        out = toggle_host_bits(nm, "192.168.0.0", toggle=1)
        self.assertEqual(out, "192.168.255.255")

    async def test_ip_from_last(self):
        # Last 255 IP is broadcast and can't be used.
        out = ip_from_last(2, "255.255.255.0", "192.168.21.0")
        self.assertEqual("192.168.21.253", out)
        out = ip_from_last(1, "255.255.255.0", "192.168.21.0")
        self.assertEqual("192.168.21.254", out)

    async def test_ipv6_rand_host(self):
        out = ipv6_rand_host("2402:1f00:8101:83f0::0000", 64)
        self.assertTrue("2402:1f00:8101:83f0" in out)
        x = ipv6_norm(out)

    async def test_gen_mac(self):
        m = generate_mac()
        b = mac_to_b(m)

    """
    TODO: 
    async def test_nt_net(self):
        if platform.system() == "Windows":
            out = await nt_ipconfig()
            print(out)

            out = await nt_route_print(desc=None)
            print(out)
    """



if __name__ == '__main__':
    main()


from p2pd import *

TOXID_HOST = "127.0.0.1"
TOXID_PORT = 8867

async def example():
    # Start the Toxiproxy server.
    i = await Interface().start_local()
    toxid = ToxiMainServer([i])
    route = await i.route().bind(port=TOXID_PORT, ips=TOXID_HOST)
    await toxid.listen_specific(
        [[route, TCP]]
    )
    #
    # Create a Toxiproxy client.
    # This just connects to the server above.
    toxid_addr = await Address(TOXID_HOST, TOXID_PORT, route)
    client = await ToxiClient(toxid_addr).start()
    #
    # Create a new relay to an upstream.
    # The upstream location is set bellow.
    relay_dest = await Address("example.com", 80, route)
    tunnel = await client.new_tunnel(relay_dest)
    #
    # Add an upstream 'toxic' that adds latency to replies.
    toxic = ToxiToxic().upstream().add_latency(ms=2000, jitter=200)
    await tunnel.new_toxic(toxic)
    #
    # Get a pipe to the tunnel (you're the 'downstream.')
    # Send data down the tunnel to the upstream.
    pipe, tunnel_tup = await tunnel.get_pipe()
    await pipe.send(b"some malformed http req\r\n\r\n")
    #
    # ... response in 2 sec +/- jitter ms.
    # If you care to do a recv here...
    #
    # Cleanup everything gracefully.
    await pipe.close()
    await tunnel.close()
    await toxid.close()
    #
    # Give tasks time to finish..
    await asyncio.sleep(2)

if __name__ == '__main__':
    async_test(example)


from p2pd.test_init import *
from p2pd import IPRange, HOST_TYPE_IP, HOST_TYPE_DOMAIN
from p2pd import IP6, IP4, Route
from p2pd.interface import Interface
from p2pd.net import Bind, cidr_to_netmask
from p2pd.route import *

class TestRoute(unittest.IsolatedAsyncioTestCase):
    async def test_invalid_wan(self):
        def ambiguous_af_ip():
            IPRange(0)

        # If zero is passed do you want a V4 or V5 -- the netmask tells which.
        self.assertRaises(Exception, ambiguous_af_ip)

        # Check netmask works for V4.
        ipr = IPRange(0, netmask="255.255.255.255")
        self.assertEqual(len(ipr), 1)
        self.assertEqual(str(ipr[0]), "0.0.0.0")

        # Check netmask works for V6.
        ip = "FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF"
        ipr = IPRange(0, netmask=ip)
        self.assertEqual(str(ipr[0]), "::")

        # Test invalid WAN is detected.
        ipr_nic = IPRange("192.168.0.20")
        def invalid_wan():
            ipr_wan = IPRange(0, "255.255.255.255")
            route = Route(IP4, [ipr_nic], [ipr_wan])

        self.assertRaises(Exception, invalid_wan)

    async def test_bind_to_route(self):
        i = await Interface().start_local()
        b = await Bind(i, i.supported()[0]).bind()
        r = await bind_to_route(b)

    async def test_if_to_rp(self):
        i = await Interface().start_local()
        rp = interfaces_to_rp([i])

    async def test_netiface_addr_to_ipr(self):
        loop = asyncio.get_event_loop()
        i = await Interface().start_local()
        af = i.supported()[0]
        r = i.route(af)
        info = {
            "addr": str(r.nic_ips[0]),
            "netmask": cidr_to_netmask(r.nic_ips[0].cidr, af)
        }

        ipr = await netiface_addr_to_ipr(af, info, i, loop, skip_bind_test=0)
        self.assertTrue(ipr is not None)


    async def test_get_routes(self):
        await init_p2pd()
        """
        * public nic IPR
            * pub nic = wan ip case
            * pub nic != wan ip case
            --------------------------
            * skip resolve case

        """
        af = IP4
        interface = None
        net_ifaces = FakeNetifaces()
        stun_client = FakeSTUNClient(interface, af=af)
        loop = asyncio.get_event_loop()

        ############################################################
        # Case: pub nic = wan ip
        ip_str = "8.8.8.8"
        net_ifaces.set_addr_info(
            {
                2: [
                    {'addr': ip_str, 'netmask': '255.255.255.255'}
                ]
            }
        )
        stun_client.set_wan_ip(ip_str)
        ret = await get_routes(
            interface=interface,
            af=af,
            netifaces=net_ifaces,
            stun_client=stun_client,
            skip_bind_test=1
        )

        # Should return a route with the same external address as the nic.
        routes, link_local = ret
        route = routes[0]
        expect_ext = IPRange(ip_str)
        self.assertEqual(route.ext_ips[0], expect_ext)
        self.assertEqual(route.nic_ips[0], expect_ext)
        self.assertTrue(route.ext_ips[0].is_public)
        self.assertTrue(route.nic_ips[0].is_public)

        # Netriface addr should not be none.
        nic_ipr = await netiface_addr_to_ipr(af, net_ifaces.addr_info[af][0], interface, loop, skip_bind_test=1)
        self.assertTrue(nic_ipr is not None)

        ###################################################################
        # Case: pub nic != wan ip
        ip_str = "7.7.7.7"
        stun_client.set_wan_ip(ip_str)
        routes, link_local = await get_routes(
            interface=interface,
            af=af,
            netifaces=net_ifaces,
            stun_client=stun_client,
            skip_bind_test=1
        )

        # Correct IP shouild be set and NIC IP marked private.
        route = routes[0]
        expect_ext = IPRange(ip_str)
        self.assertEqual(route.nic_ips[0], nic_ipr)
        self.assertEqual(route.ext_ips[0], expect_ext)
        self.assertTrue(route.ext_ips[0].is_public)
        self.assertFalse(route.nic_ips[0].is_public)

        ###################################################################
        # Case: pub nic != wan ip; skip_resolve = True
        routes, link_local = await get_routes(
            interface=interface,
            af=af,
            netifaces=net_ifaces,
            stun_client=stun_client,
            skip_bind_test=1,
            skip_resolve=True
        )

        # Should pretend pub nic = ext when res is disabled.
        route = routes[0]
        self.assertEqual(route.nic_ips[0], nic_ipr)
        self.assertEqual(route.ext_ips[0], nic_ipr)
        self.assertTrue(route.ext_ips[0].is_public)
        self.assertTrue(route.nic_ips[0].is_public)

    async def test_route_ops(self):
        r1 = Route(
            af=IP4,
            nic_ips=[IPRange("192.168.0.1")],
            ext_ips=[IPRange("8.8.8.8")]
        )
        self.assertEqual(len(r1), 1)
        self.assertEqual(r1, r1)

        r2 = Route(
            af=IP4,
            nic_ips=[IPRange("192.168.0.10")],
            ext_ips=[IPRange("8.8.8.8")]
        )

        r3 = Route(
            af=IP4,
            nic_ips=[IPRange("192.168.0.10")],
            ext_ips=[IPRange("8.8.8.10")]
        )
        self.assertNotEqual(r1, r3  )

        r4 = Route(
            af=IP4,
            nic_ips=[IPRange("192.168.0.10")],
            ext_ips=[IPRange("8.8.8.12")]
        )

        r5 = Route(
            af=IP6,
            nic_ips=[IPRange("2402:1f00:8101:83f::1")],
            ext_ips=[IPRange("2402:1f00:8101:83f::1")]
        )

        rp = RoutePool([r1, r2, r3, r4, r5])
        nic_ipr_a = IPRange("192.168.0.2")
        nic_ipr_b = IPRange("192.168.0.1")
        ipr_c = IPRange("8.8.8.8")
        ipr_d = IPRange("2402:1f00:8101:83f::2")

        # Can find NIC properly.
        self.assertFalse(r1.has_nic_ip(nic_ipr_a))
        self.assertTrue(r1.has_nic_ip(nic_ipr_b))

        # Test equality (same ext = same route)
        self.assertEqual(r1, r2)
        self.assertTrue(r1 in r2)

        # Test invert (find route with different wan ip).
        r = ~r1
        self.assertTrue(r.ext_ips[0] != ipr_c)
        
        # Test not equal.
        self.assertTrue(r != r1)

        # Get N alternative routes that use a different IP.
        # Exclusions are supported.
        exclusions = [r3]
        routes = r1.alt(limit=10, exclusions=exclusions)
        self.assertTrue(len(routes))
        self.assertFalse(r1 in routes)
        self.assertFalse(r3 in routes)

        # Check all different conversions.
        self.assertFalse(r4 == ipr_c) # IPRange
        self.assertFalse(r4 == "10.0.0.1") # Str
        self.assertFalse(r4 == b"10.0.0.1") # bytes
        self.assertFalse(r5 == int(ipr_d)) # int
        self.assertFalse(r4 == ipaddress.IPv4Address(1)) # IPaddress

        # Check route pool iter code.
        i = 0
        for route in rp:
            self.assertTrue(isinstance(route, Route))
            i += 1
            if i >= 2:
                break

        # Check reversed works.
        i = 0
        for route in reversed(rp):
            self.assertTrue(isinstance(route, Route))
            i += 1
            if i >= 2:
                break

    async def test_route_pool(self):
        # 511 hosts overall
        af = IP4
        ipr_a = IPRange("8.8.8.0", cidr=24) # 255 hosts
        ipr_b = IPRange("7.7.7.0", cidr=24) # 255 hosts
        ipr_c = IPRange("9.9.9.9") # 1 host
        
        # Setup route pool
        r1 = Route(af, [ipr_a], [copy.deepcopy(ipr_a)])
        r2 = Route(af, [ipr_b], [copy.deepcopy(ipr_b)])
        r3 = Route(af, [ipr_c], [copy.deepcopy(ipr_c)])
        rp = RoutePool([r1, r2, r3])

        # Test dict serialization.
        d = rp.to_dict()
        rp = RoutePool.from_dict(d)

        # block a + block b + 1 host = 511
        self.assertEqual(rp.wan_hosts, 511)
        self.assertEqual(len(rp), 511)

        # Index a host into first range.
        ipr_x = IPRange("8.8.8.11")
        self.assertEqual(rp[10], ipr_x)

        # Index a host into second range.
        ipr_x = IPRange("7.7.7.11")
        self.assertEqual(rp[265], ipr_x)

        # Try index last host in range using negative index.
        self.assertEqual(rp[-1], ipr_c)

        # Test slicing between ranges.
        x = rp[254:256]
        self.assertEqual(len(x), 2)
        self.assertEqual(x[0].ext_ips[0], IPRange("8.8.8.255"))
        self.assertEqual(x[1].ext_ips[0], IPRange("7.7.7.1")) 

        # Test tuple list fetching.
        x = rp[(254, 255)]
        self.assertEqual(len(x), 2)
        self.assertEqual(x[0].ext_ips[0], IPRange("8.8.8.255"))
        self.assertEqual(x[1].ext_ips[0], IPRange("7.7.7.1")) 

        # Test reversed iter.
        for r in reversed(rp):
            self.assertEqual(r.ext_ips[0], ipr_c)
            break

        # Test regular iter.
        for r in rp:
            self.assertEqual(r.ext_ips[0], IPRange("8.8.8.1"))
            break

        # Test pop.
        x = rp.pop()
        self.assertEqual(x.ext_ips[0], IPRange("8.8.8.1"))
        x = rp.pop()
        self.assertEqual(x.ext_ips[0], IPRange("8.8.8.2"))

if __name__ == '__main__':
    main()


from p2pd.test_init import *
from p2pd import Address, HOST_TYPE_IP, HOST_TYPE_DOMAIN
from p2pd import IP6, IP4
from p2pd import Interface

class TestIPRange(unittest.IsolatedAsyncioTestCase):
    async def test_domain_resolve(self):
        i = await Interface().start_local()
        r = i.route()
        a = await Address("www.google.com", 80, r).res()
        self.assertEqual(a.host_type, HOST_TYPE_DOMAIN)

    async def test_v6_resolve(self):
        i = await Interface().start_local()
        ip = "2402:1f00:8101:083f:0000:0000:0000:0001"
        r = i.route(IP6)
        a = await Address(ip, 80, r).res()
        self.assertEqual(a.host_type, HOST_TYPE_IP)
        self.assertTrue(a.is_public)

    async def test_v4_resolve(self):
        i = await Interface().start_local()
        r = i.route(IP4)
        ip = "192.168.0.1"
        a = await Address(ip, 80, r).res()
        self.assertEqual(a.host_type, HOST_TYPE_IP)
        self.assertTrue(a.is_private)

if __name__ == '__main__':
    main()